//wd1772_real.c
// management of WD1772 data, real chip

#include <stdio.h>
#include "wd1772_real.h"

namespace xatremu
{

WD1772Real::WD1772Real(){
	m_trackData = new unsigned char[0x4000];
	m_ATRData = new unsigned char[368272+16];
	ResetTrackWriteProperties(true);
#ifdef WITH_CAPS
	m_CAPSImgPresent = false;
#endif
	m_WATRHeader.DiskSectorLen = 0x80;
	m_WATRHeader.SidePtr[0] = NULL;
	m_WATRHeader.SidePtr[1] = NULL;
};

WD1772Real::~WD1772Real(){
	if(m_ATRData != NULL) {
		delete[] m_ATRData;
		m_ATRData = NULL;
	};
	if(m_trackData != NULL) {
		delete[] m_trackData;
		m_trackData = NULL;
	};
	ResetTrackWriteProperties(false);
	//free WATR buffer
	FreeWATRData();
};

void WD1772Real::ResetWD1772(){
	m_MotorIsOn = false;
	m_CurrentTrack = 2;
	m_imageType = 1;
	m_encodingType = 0;
	m_imageLoaded = false;
	m_Is288RPM = false;
	m_WriteProtect = false;
	m_WD1772State = wd1772_idle;
};

void WD1772Real::ChooseDoubleDensity(bool doubleDensity){
#ifdef TESTSTATE
	printf("Double Density is %d\n", doubleDensity);
#endif
	m_IsDoubleDensity = doubleDensity;
};

void WD1772Real::MasterReset(){
	m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772MOTOR, MAXCYCLE);
	m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772HEADSTEP, MAXCYCLE);
	m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772HEADCALM, MAXCYCLE);
	m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, MAXCYCLE);
	m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772DATATIMEOUT, MAXCYCLE);
	MotorOff();	
};

void WD1772Real::ChooseRegister(unsigned char reg){
	m_SelectedRegister = reg & 3;
};

void WD1772Real::ChooseReadRegisters(bool readRegs){
	m_SelectRegsForRead = readRegs;
};

void WD1772Real::ChooseSide(bool side){
	m_IsDiskSide1 = side;
	CreateTrackData();
};

bool WD1772Real::SetImage(char* imageName){
#ifdef TESTSTATE
	printf("SetImage entered\n");
#endif
	m_imageType = 0;
	m_imageLoaded = false;
	m_trackIsValid = false;

#ifdef WITH_CAPS
	if(m_CAPSImgPresent){
		CAPSUnlockImage(m_CAPSImgId);
		m_CAPSImgPresent = false;
	};

	//TODO: check for IPF

	//Check for CT Raw
	if(CAPSGetImageType(imageName) == citCTRaw){
		CapsImageInfo myCAPSImageInfo;
		if((CAPSLockImage(m_CAPSImgId, imageName) == imgeOk) &&
				CAPSGetImageInfo(&myCAPSImageInfo, m_CAPSImgId) == imgeOk){
			if((myCAPSImageInfo.mincylinder==0) && (myCAPSImageInfo.maxcylinder>=78)){
#ifdef TESTSTATE3
	printf("CT image found\n");
#endif
				m_CAPSMinHead = myCAPSImageInfo.minhead;
				m_CAPSMaxHead = myCAPSImageInfo.maxhead;

				//copy image name
				strncpy(&m_GivenFileName[0], imageName, 520);
				m_GivenFileName[520] = 0;
				m_GivenFileName[521] = 0;
				m_GivenFileName[522] = 0;
				m_GivenFileName[523] = 0;
				m_GivenFileName[524] = 0;
				strncpy(&m_WATRFileName[0], &m_GivenFileName[0], 524);
				bool finished = false;
				unsigned short int pointPos = strlen(m_WATRFileName);
				unsigned short checkPos = strlen(m_WATRFileName);
				while (!finished){
					if(checkPos == 0){
						finished = true;
					};
					--checkPos;
					if(checkPos == 0){
						finished = true;
					}
					else{
						if(m_WATRFileName[checkPos] == '.'){
							pointPos = checkPos;
							finished = true;
						};
						if((m_WATRFileName[checkPos] == '/') || (m_WATRFileName[checkPos] == '\\')){
							finished = true;
						};
					};
				};
				m_WATRFileName[pointPos] = '.';
				m_WATRFileName[pointPos+1] = 'W';
				m_WATRFileName[pointPos+2] = 'A';
				m_WATRFileName[pointPos+3] = 'T';
				m_WATRFileName[pointPos+4] = 'R';
				m_WATRFileName[pointPos+5] = 0;

				//load WATR data
				FreeWATRData();
				LoadWATRData();

				//load 5 revolutions
				for(int iloop = 0; iloop < 5; ++iloop){
					if(CAPSLockTrack(&m_CAPSTrackInfo[iloop], m_CAPSImgId, m_CurrentTrack*2, 
							(m_IsDiskSide1) && (m_CAPSMaxHead > m_CAPSMinHead) ? m_CAPSMinHead+1 : m_CAPSMinHead,
							DI_LOCK_INDEX|DI_LOCK_NOISE|DI_LOCK_TRKBIT|DI_LOCK_UPDATEFD) != imgeOk) throw;
				};
				m_imageType = 2;
				m_CurrentRevolution=0;

#ifdef TESTSTATE3
	printf("CT image loaded\n");
#endif
				ResetTrackWriteProperties(false);
				//load trackData from internal WATR buffer
				TransferTrackWriteProperties();

				//analyze track for format and copy data into buffer
				m_encodingType = AnalyzeFromCT();
				m_imageLoaded = true;
				return true;
			};
		};
		return false;
	};
#endif

	//check for ATR
	FILE * f;
	f = fopen(imageName, "rb");
	if (f == NULL)
		return false;
	fseek(f, 0L, SEEK_END);
	long int len = ftell(f);
	if (len > 368272){
		fclose(f);
		return false;
	};
	fseek(f, 0L, SEEK_SET);
	fread(m_ATRData, 1, 368272, f);
	fclose(f);

	//copy image name
	strncpy(&m_GivenFileName[0], imageName, 520);
	m_GivenFileName[520] = 0;
	m_GivenFileName[521] = 0;
	m_GivenFileName[522] = 0;
	m_GivenFileName[523] = 0;
	m_GivenFileName[524] = 0;
	strncpy(&m_WATRFileName[0], &m_GivenFileName[0], 524);
	bool finished = false;
	unsigned short int pointPos = strlen(m_WATRFileName);
	unsigned short checkPos = strlen(m_WATRFileName);
	while (!finished){
		if(checkPos == 0){
			finished = true;
		};
		--checkPos;
		if(checkPos == 0){
			finished = true;
		}
		else{
			if(m_WATRFileName[checkPos] == '.'){
				pointPos = checkPos;
				finished = true;
			};
			if((m_WATRFileName[checkPos] == '/') || (m_WATRFileName[checkPos] == '\\')){
				finished = true;
			};
		};
	};
	m_WATRFileName[pointPos] = '.';
	m_WATRFileName[pointPos+1] = 'W';
	m_WATRFileName[pointPos+2] = 'A';
	m_WATRFileName[pointPos+3] = 'T';
	m_WATRFileName[pointPos+4] = 'R';
	m_WATRFileName[pointPos+5] = 0;

#ifdef TESTSTATE
	printf("WATR Filename: %s\n", m_WATRFileName);
#endif

	if((m_ATRData[0] == 0x96) &&
		(m_ATRData[1] == 0x02) &&
		(m_ATRData[2] == 0x80) &&
		(m_ATRData[3] == 0x16) &&
		(m_ATRData[4] == 0x80) &&
		(m_ATRData[5] == 0x00)){
		m_imageLoaded = true;
		m_imageType = 1;
		m_encodingType = 0;
		//load WATR data
		FreeWATRData();
		LoadWATRData();
#ifdef TESTSTATE
		printf("ATR Single image detected\n");
#endif
		CreateTrackData();
		return true;
	};

	if((m_ATRData[0] == 0x96) &&
		(m_ATRData[1] == 0x02) &&
		(m_ATRData[2] == 0x80) &&
		(m_ATRData[3] == 0x20) &&
		(m_ATRData[4] == 0x80) &&
		(m_ATRData[5] == 0x00)){
		m_imageLoaded = true;
		m_imageType = 1;
		m_encodingType = 1;
		//load WATR data
		FreeWATRData();
		LoadWATRData();
#ifdef TESTSTATE
		printf("ATR Medium image detected\n");
#endif
		CreateTrackData();
		return true;
	};

	if((m_ATRData[0] == 0x96) &&
		(m_ATRData[1] == 0x02) &&
		(m_ATRData[2] == 0xe8) &&
		(m_ATRData[3] == 0x2c) &&
		(m_ATRData[4] == 0x00) &&
		(m_ATRData[5] == 0x01)){
		m_imageLoaded = true;
		m_imageType = 1;
		m_encodingType = 2;
		//load WATR data
		FreeWATRData();
		LoadWATRData();
#ifdef TESTSTATE
		printf("ATR Double image detected\n");
#endif
		CreateTrackData();
		return true;
	};

	if((m_ATRData[0] == 0x96) &&
		(m_ATRData[1] == 0x02) &&
		(m_ATRData[2] == 0xe8) &&
		(m_ATRData[3] == 0x59) &&
		(m_ATRData[4] == 0x00) &&
		(m_ATRData[5] == 0x01)){
		m_imageLoaded = true;
		m_imageType = 1;
		m_encodingType = 3;
		//load WATR data
		FreeWATRData();
		LoadWATRData();
#ifdef TESTSTATE
		printf("ATR Quad image detected\n");
#endif
		CreateTrackData();
		return true;
	};

	return false;
};

void WD1772Real::EjectImage(){
	m_imageType = 0;
	m_imageLoaded = false;
	m_trackIsValid = false;
};

void WD1772Real::Choose288RPM(bool rpm288){
	m_Is288RPM = rpm288;
};

bool WD1772Real::GetIntRq(){
	bool tmpReq = m_IntRQ;
	m_IntRQ = false;
	return tmpReq;
};

//data request line
bool WD1772Real::StateT1Line(){
	bool tmp = m_DRQ;
	m_DRQ = false;
	return tmp;
};

//check if next drq is due and count the byte offset on the track
void WD1772Real::CheckDRQEvent(){
	switch(m_WD1772State){
	case wd1772_readsearchheadersync:
		//if fm and not double density selected
		if(m_trackIsValid && (m_encodingType == 0) && !m_IsDoubleDensity) {
			//look if next trackdataword is sync of sector header
			unsigned long int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
#ifdef TESTSTATE
//			printf("TD is %8x at %4x\n", tmpDataLong, m_currentTrackOffset);
#endif
			m_SyncBitShift = SearchSync(tmpDataLong, 0xf57e, 0xf57e); //FE
			if(m_SyncBitShift >= 0){
				m_WD1772State = wd1772_readgotheadersync;
				m_headerData[0] = 0xfe;
				m_WordsElapsed = 1;
				PrepareMotorOff();
				m_LastHeaderOffset = m_currentTrackOffset;
#ifdef TESTSTATE
				printf("Sync Shift is %d\n", m_SyncBitShift);
#endif
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		//if mfm medium and double density selected
		if(m_trackIsValid && (m_encodingType >= 1) && (m_encodingType <= 3) && m_IsDoubleDensity) {
			//look if next trackdataword is sync of sector header
			unsigned long int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
#ifdef TESTSTATE
//			printf("TD is %8x at %4x\n", tmpDataLong, m_currentTrackOffset);
#endif
			m_SyncBitShift = SearchSync(tmpDataLong, 0x4489, 0x4489);
			if(m_SyncBitShift >= 0){
				m_LastHeaderOffset = m_currentTrackOffset;
				m_WD1772State = wd1772_readgotgenericsyncforheader;
				PrepareMotorOff();
#ifdef TESTSTATE
				printf("Sync Shift is %d\n", m_SyncBitShift);
#endif
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		break;

	case wd1772_readgotgenericsyncforheader:
		//if mfm medium and double density selected
		if(m_trackIsValid && (m_encodingType >= 1) && (m_encodingType <= 3) && m_IsDoubleDensity) {
			//look if next trackdataword is sync of sector header
			unsigned long int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
#ifdef TESTSTATE
			printf("TD is %8x at %4x\n", tmpDataLong, m_currentTrackOffset);
#endif
			tmpDataLong = tmpDataLong >> (16 - m_SyncBitShift);
			if((tmpDataLong & 0xffff) == 0x4489){
				//do nothing
			}
			else {
				m_headerData[0] = DecodeByteMFM(tmpDataLong & 0xffff);
				if(m_headerData[0] == 0xfe){
					m_WD1772State = wd1772_readgotheadersync;
					m_WordsElapsed = 1;
				}
				else {
					m_WD1772State = wd1772_readsearchheadersync;
				};
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		break;

	case wd1772_readgotheadersync:
		//if fm and not double density selected
		if(m_trackIsValid && (m_encodingType == 0) && !m_IsDoubleDensity) {
			unsigned int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
			tmpDataLong = tmpDataLong >> (16 - m_SyncBitShift);
			m_headerData[m_WordsElapsed] = DecodeByteFM(tmpDataLong & 0xffff);
#ifdef TESTSTATE
			printf("Header data %2x at %2x\n", m_headerData[m_WordsElapsed], m_WordsElapsed);
#endif
			++m_WordsElapsed;
			if(m_WordsElapsed == 7){
				m_CurrentCRC = 0xffff;
				for(int loop = 0; loop < 5; ++loop) {
					m_CurrentCRC = CalculateCRC(m_CurrentCRC, m_headerData[loop]);
				};
				m_CurrentCRC = CalculateCRCEnd(m_CurrentCRC);
				unsigned short int diskCRC = (m_headerData[5] << 8) + m_headerData[6];
#ifdef TESTSTATE3
//				if(m_headerData[3] == 2){
//					printf("header data: sector %1x, track %1x, side %1x, len %1x, CTOffset %4x\n", m_headerData[3], m_headerData[1], m_headerData[2], m_headerData[4], m_currentTrackOffset);
//				};
#endif
				if(m_CurrentTrack != m_headerData[1]){
					m_WD1772State = wd1772_readsearchheadersync;
				}
				else if(m_WantedSector != m_headerData[3]){
					m_WD1772State = wd1772_readsearchheadersync;
				}
				else {
					m_SectorLength = 0x80 << m_headerData[4];
					if(m_SectorLength > 2048)
						m_SectorLength = 2048;
					if(m_CurrentCRC != diskCRC){
						m_CRCError = true;
						m_WD1772State = wd1772_readsearchheadersync;
#ifdef TESTSTATE3
					printf("header CRC error for sector %d, track %d\n", m_WantedSector, m_WantedTrack);
#endif
					}
					else{
						m_CRCError = false;
						bool found = false;
						//search entry in sectorwritetable
						short int sequence = 0;
						while(!found && (sequence < 30) 
							&& (m_SectorLayout[m_headerData[3]-1][sequence].SectorSequenceNumber != 0xff)){
							unsigned short tmpOffset = m_SectorLayout[m_headerData[3]-1][sequence].OffsetInProcessed;

							if(m_LastHeaderOffset < 30){
								if((m_LastHeaderOffset < (tmpOffset + 30)) ||
									(m_LastHeaderOffset > (m_EncodedBytesPerTrack + tmpOffset - 30))){
									m_CurrentSectorSequenceNumber = sequence;
									found = true;
								};
							}
							else if(m_LastHeaderOffset > (m_EncodedBytesPerTrack - 30)){
								if((m_LastHeaderOffset < (tmpOffset + 30 - m_EncodedBytesPerTrack)) || 
									(m_LastHeaderOffset > (tmpOffset - 30))){
									m_CurrentSectorSequenceNumber = sequence;
									found = true;
								};
							}
							else{
								if((m_LastHeaderOffset < (tmpOffset + 30)) &&
									(m_LastHeaderOffset > (tmpOffset - 30))){
									m_CurrentSectorSequenceNumber = sequence;
									found = true;
								};
							};

							++sequence;
						};
						if((m_LastCommand & 0xf0) == 0xa0){
							if(found){
								if(m_SectorLayout[m_headerData[3]-1][m_CurrentSectorSequenceNumber].SectorData == NULL){
									m_SectorLayout[m_headerData[3]-1][m_CurrentSectorSequenceNumber].SectorData = new unsigned char[m_SectorLengthInTrack];
								};
								m_SectorDataOnWrite = m_SectorLayout[m_headerData[3]-1][m_CurrentSectorSequenceNumber].SectorData;
							}
							else{
								m_SectorDataOnWrite = &m_SectorWriteDummy[0];
							};
							//next write state
							m_WD1772State = wd1772_readuntilwritedata;
						}
						else{
							if(found){
								m_SectorDataOnWrite = m_SectorLayout[m_headerData[3]-1][m_CurrentSectorSequenceNumber].SectorData;
							}
							else{
								m_SectorDataOnWrite = NULL;
							};
							m_WD1772State = wd1772_readsearchdatasync;
						};
						m_WordsElapsed = 0;
#ifdef TESTSTATE3
					printf("header found for sector %d, track %d, CurrentTrackOffset %4x\n", m_WantedSector, m_WantedTrack, m_currentTrackOffset);
#endif
					};
				};
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		//if mfm medium and double density selected
		if(m_trackIsValid && (m_encodingType >= 1) && (m_encodingType <= 3) && m_IsDoubleDensity) {
			unsigned int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
			tmpDataLong = tmpDataLong >> (16 - m_SyncBitShift);
			m_headerData[m_WordsElapsed] = DecodeByteMFM(tmpDataLong & 0xffff);
#ifdef TESTSTATE
			printf("Header data %2x at %2x\n", m_headerData[m_WordsElapsed], m_WordsElapsed);
#endif
			++m_WordsElapsed;
			if(m_WordsElapsed == 7){
				m_CurrentCRC = 0x2db3;
				for(int loop = 0; loop < 5; ++loop) {
					m_CurrentCRC = CalculateCRC(m_CurrentCRC, m_headerData[loop]);
				};
				m_CurrentCRC = CalculateCRCEnd(m_CurrentCRC);
				unsigned short int diskCRC = (m_headerData[5] << 8) + m_headerData[6];
				if(m_CurrentTrack != m_headerData[1]){
					m_WD1772State = wd1772_readsearchheadersync;
				}
				else if(m_WantedSector != m_headerData[3]){
					m_WD1772State = wd1772_readsearchheadersync;
				}
				else {
					m_SectorLength = 0x80 << m_headerData[4];
					if(m_SectorLength > 2048)
						m_SectorLength = 2048;
					if(m_CurrentCRC != diskCRC){
						m_CRCError = true;
						m_WD1772State = wd1772_readsearchheadersync;
					}
					else{
						m_CRCError = false;
						bool found = false;
						//search entry in sectorwritetable
						short int sequence = 0;
						while(!found && (sequence < 30) 
							&& (m_SectorLayout[m_headerData[3]-1][sequence].SectorSequenceNumber != 0xff)){
							unsigned short tmpOffset = m_SectorLayout[m_headerData[3]-1][sequence].OffsetInProcessed;

							if(m_LastHeaderOffset < 30){
								if((m_LastHeaderOffset < (tmpOffset + 30)) ||
									(m_LastHeaderOffset > (m_EncodedBytesPerTrack + tmpOffset - 30))){
									m_CurrentSectorSequenceNumber = sequence;
									found = true;
								};
							}
							else if(m_LastHeaderOffset > (m_EncodedBytesPerTrack - 30)){
								if((m_LastHeaderOffset < (tmpOffset + 30 - m_EncodedBytesPerTrack)) || 
									(m_LastHeaderOffset > (tmpOffset - 30))){
									m_CurrentSectorSequenceNumber = sequence;
									found = true;
								};
							}
							else{
								if((m_LastHeaderOffset < (tmpOffset + 30)) &&
									(m_LastHeaderOffset > (tmpOffset - 30))){
									m_CurrentSectorSequenceNumber = sequence;
									found = true;
								};
							};

							++sequence;
						};
						if((m_LastCommand & 0xf0) == 0xa0){
							if(found){
								if(m_SectorLayout[m_headerData[3]-1][m_CurrentSectorSequenceNumber].SectorData == NULL){
									m_SectorLayout[m_headerData[3]-1][m_CurrentSectorSequenceNumber].SectorData = new unsigned char[m_SectorLengthInTrack];
								};
								m_SectorDataOnWrite = m_SectorLayout[m_headerData[3]-1][m_CurrentSectorSequenceNumber].SectorData;
							}
							else{
								m_SectorDataOnWrite = &m_SectorWriteDummy[0];
							};
							//next write state
							m_WD1772State = wd1772_readuntilwritedata;
						}
						else{
							if(found){
								m_SectorDataOnWrite = m_SectorLayout[m_headerData[3]-1][m_CurrentSectorSequenceNumber].SectorData;
							}
							else{
								m_SectorDataOnWrite = NULL;
							};
							m_WD1772State = wd1772_readsearchdatasync;
						};
						m_WordsElapsed = 0;
					};
				};
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		break;

	case wd1772_readsearchdatasync:
		//must be found within 30 words
		//if fm and not double density selected
		if(m_trackIsValid && (m_encodingType == 0) && !m_IsDoubleDensity) {
			//look if next trackdataword is sync of sector data
			unsigned long int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
			m_SyncBitShift = SearchSync(tmpDataLong, 0xf56f, 0xf56a); //FB, F8
			++m_WordsElapsed;
			if(m_SyncBitShift >= 0){
				m_WD1772State = wd1772_readgetdata;
				unsigned int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
				tmpDataLong = tmpDataLong >> (16 - m_SyncBitShift);
				m_headerData[0] = (m_SectorDataOnWrite != NULL) ? 0xFB : DecodeByteFM(tmpDataLong & 0xffff);
				m_CurrentCRC = CalculateCRC(0xffff, m_headerData[0]);
				m_WordsElapsed = 0;
			}
			else if (m_WordsElapsed >= 30){
				m_WD1772State = wd1772_readsearchheadersync;
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		//must be found within 43 words
		//if mfm and double density selected
		if(m_trackIsValid && (m_encodingType >= 1) && (m_encodingType <= 3)  && m_IsDoubleDensity) {
			//look if next trackdataword is sync of sector data
			unsigned long int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
			m_SyncBitShift = SearchSync(tmpDataLong, 0x4489, 0x4489);
			++m_WordsElapsed;
			if(m_SyncBitShift >= 0){
				m_WD1772State = wd1772_readgotdatasync;
				m_WordsElapsed = 0;
			}
			else if (m_WordsElapsed >= 43){
				m_WD1772State = wd1772_readsearchheadersync;
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		break;

	case wd1772_readgotdatasync:
		//if mfm medium and double density selected
		if(m_trackIsValid && (m_encodingType >= 1) && (m_encodingType <= 3)  && m_IsDoubleDensity) {
			//look if next trackdataword is sync of sector header
			unsigned long int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
#ifdef TESTSTATE
//			printf("TD is %8x at %4x\n", tmpDataLong, m_currentTrackOffset);
#endif
			tmpDataLong = tmpDataLong >> (16 - m_SyncBitShift);
			if((tmpDataLong & 0xffff) == 0x4489){
				//do nothing
			}
			else {
				unsigned char tmpValue = DecodeByteMFM(tmpDataLong & 0xffff);
				if((tmpValue == 0xfb) || (tmpValue == 0xf8)){
					m_headerData[0] = (m_SectorDataOnWrite != NULL) ? 0xFB : tmpValue;
					m_WD1772State = wd1772_readgetdata;
					m_CurrentCRC = CalculateCRC(0x2db3, m_headerData[0]);
					m_WordsElapsed = 0;
				}
				else {
					m_WD1772State = wd1772_readsearchheadersync;
				};
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		break;

	case wd1772_readgetdata:
		//if fm and not double density selected
		if(m_trackIsValid && (m_encodingType == 0) && !m_IsDoubleDensity) {
			if(m_SectorDataOnWrite == NULL){
				unsigned int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
				tmpDataLong = tmpDataLong >> (16 - m_SyncBitShift);
				m_CurrentByte = DecodeByteFM(tmpDataLong & 0xffff);
			}
			else{
				if(m_WordsElapsed < m_SectorLengthInTrack){
					unsigned char * tmpPtr = m_SectorDataOnWrite + m_WordsElapsed;
					m_CurrentByte = *tmpPtr;
				}
				else{
					m_CurrentByte = 0;
				};
			};
			if(m_DataByteProvided)
				m_DataOverflow = true;
			m_DataByteProvided = true;

			m_CurrentCRC = CalculateCRC(m_CurrentCRC, m_CurrentByte);

			//set DRQ
			m_DRQ = true;

			++m_WordsElapsed;
			if(m_WordsElapsed == m_SectorLength){
				m_CurrentCRC = CalculateCRCEnd(m_CurrentCRC);
				m_WD1772State = wd1772_readcheckdatacrc;
				m_WordsElapsed = 0;
			};

			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());

		};
		//if mfm and double density selected
		if(m_trackIsValid && (m_encodingType >= 1) && (m_encodingType <= 3)  && m_IsDoubleDensity) {
			if(m_SectorDataOnWrite == NULL){
				unsigned int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
				tmpDataLong = tmpDataLong >> (16 - m_SyncBitShift);
				m_CurrentByte = DecodeByteMFM(tmpDataLong & 0xffff);
			}
			else{
				if(m_WordsElapsed < m_SectorLengthInTrack){
					unsigned char * tmpPtr = m_SectorDataOnWrite + m_WordsElapsed;
					m_CurrentByte = *tmpPtr;
				}
				else{
					m_CurrentByte = 0;
				};
			};
			if(m_DataByteProvided)
				m_DataOverflow = true;
			m_DataByteProvided = true;

#ifdef TESTSTATE2
    if (m_CurrentTrack == 39)
    {
        printf("Data of Side %d, Track 39, Sector %d = %d\n", (int) m_IsDiskSide1, m_WantedSector, m_CurrentByte);
    };
#endif
 
			m_CurrentCRC = CalculateCRC(m_CurrentCRC, m_CurrentByte);

			//set DRQ
			m_DRQ = true;

			++m_WordsElapsed;
			if(m_WordsElapsed == m_SectorLength){
				m_CurrentCRC = CalculateCRCEnd(m_CurrentCRC);
				m_WD1772State = wd1772_readcheckdatacrc;
				m_WordsElapsed = 0;
			};

			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());

		};
		break;

	case wd1772_readcheckdatacrc:
		//if fm and not double density selected
		if(m_trackIsValid && (m_encodingType == 0) && !m_IsDoubleDensity) {
			unsigned int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
			tmpDataLong = tmpDataLong >> (16 - m_SyncBitShift);
			m_headerData[1+m_WordsElapsed] = DecodeByteFM(tmpDataLong & 0xffff);

			++m_WordsElapsed;
			if(m_WordsElapsed == 2){
				unsigned short int diskCRC = (m_headerData[1] << 8) + m_headerData[2];
				if((m_SectorDataOnWrite == NULL) && (diskCRC != m_CurrentCRC)){
					m_CRCError = true;
#ifdef TESTSTATE3
					printf("sector CRC error for sector %d, track %d\n", m_WantedSector, m_WantedTrack);
#endif
				};
				if(m_headerData[0] == 0xF8){
					m_DeletedDataMark = true;
				};
				m_IntRQ = true;
				m_Busy = false;
				m_WD1772State = wd1772_motoridle;
				m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772DATATIMEOUT, MAXCYCLE);
				PrepareMotorOff();
			};

			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		//if mfm and double density selected
		if(m_trackIsValid && (m_encodingType >= 1) && (m_encodingType <= 3)  && m_IsDoubleDensity) {
			unsigned int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
			tmpDataLong = tmpDataLong >> (16 - m_SyncBitShift);
			m_headerData[1+m_WordsElapsed] = DecodeByteMFM(tmpDataLong & 0xffff);

			++m_WordsElapsed;
			if(m_WordsElapsed == 2){
				unsigned short int diskCRC = (m_headerData[1] << 8) + m_headerData[2];
				if((m_SectorDataOnWrite == NULL) && (diskCRC != m_CurrentCRC)){
					m_CRCError = true;
				};
				if(m_headerData[0] == 0xF8){
					m_DeletedDataMark = true;
				};
				m_IntRQ = true;
				m_Busy = false;
				m_WD1772State = wd1772_motoridle;
				m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772DATATIMEOUT, MAXCYCLE);
				PrepareMotorOff();
			};

			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		break;

	case wd1772_rasearchheadersync:
		//if fm and not double density selected
		if(m_trackIsValid && (m_encodingType == 0) && !m_IsDoubleDensity) {
			//look if next trackdataword is sync of sector header
			unsigned long int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
#ifdef TESTSTATE
//			printf("TD is %8x at %4x\n", tmpDataLong, m_currentTrackOffset);
#endif
			m_SyncBitShift = SearchSync(tmpDataLong, 0xf57e, 0xf57e); //FE
			if(m_SyncBitShift >= 0){
				m_WD1772State = wd1772_ragotheadersync;
				m_headerData[0] = 0xfe;
//				m_CurrentCRC = CalculateCRC(0xffff, m_headerData[0]);
				m_WordsElapsed = 1;
				PrepareMotorOff();
#ifdef TESTSTATE
				printf("Sync Shift is %d\n", m_SyncBitShift);
#endif
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		//if mfm medium and double density selected
		if(m_trackIsValid && (m_encodingType >= 1) && (m_encodingType <= 3)  && m_IsDoubleDensity) {
			//look if next trackdataword is sync of sector header
			unsigned long int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
#ifdef TESTSTATE
//			printf("TD is %8x at %4x\n", tmpDataLong, m_currentTrackOffset);
#endif
			m_SyncBitShift = SearchSync(tmpDataLong, 0x4489, 0x4489);
			if(m_SyncBitShift >= 0){
				m_WD1772State = wd1772_ragotgenericsyncforheader;
				//m_headerData[0] = 0xfe;
//				m_CurrentCRC = CalculateCRC(0xffff, m_headerData[0]);
				//m_WordsElapsed = 1;
				PrepareMotorOff();
#ifdef TESTSTATE
				printf("Sync Shift is %d\n", m_SyncBitShift);
#endif
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		break;

	case wd1772_ragotgenericsyncforheader:
		//if mfm medium and double density selected
		if(m_trackIsValid && (m_encodingType >= 1) && (m_encodingType <= 3)  && m_IsDoubleDensity) {
			//look if next trackdataword is sync of sector header
			unsigned long int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
#ifdef TESTSTATE
			printf("TD is %8x at %4x\n", tmpDataLong, m_currentTrackOffset);
#endif
			tmpDataLong = tmpDataLong >> (16 - m_SyncBitShift);
			if((tmpDataLong & 0xffff) == 0x4489){
				//do nothing
			}
			else {
				m_headerData[0] = DecodeByteMFM(tmpDataLong & 0xffff);
				if(m_headerData[0] == 0xfe){
					m_WD1772State = wd1772_ragotheadersync;
					m_WordsElapsed = 1;
				}
				else {
					m_WD1772State = wd1772_rasearchheadersync;
				};
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		break;

	case wd1772_ragotheadersync:
		//if fm and not double density selected
		if(m_trackIsValid && (m_encodingType == 0) && !m_IsDoubleDensity) {
			unsigned int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
			tmpDataLong = tmpDataLong >> (16 - m_SyncBitShift);
			m_CurrentByte = m_headerData[m_WordsElapsed] = DecodeByteFM(tmpDataLong & 0xffff);
			if(m_DataByteProvided)
				m_DataOverflow = true;
			m_DataByteProvided = true;
#ifdef TESTSTATE
			printf("Header data %2x at %2x\n", m_headerData[m_WordsElapsed], m_WordsElapsed);
#endif
			m_DRQ = true;

			++m_WordsElapsed;
			if(m_WordsElapsed == 7){
				m_CurrentCRC = 0xffff;
				for(int loop = 0; loop < 5; ++loop) {
					m_CurrentCRC = CalculateCRC(m_CurrentCRC, m_headerData[loop]);
				};
				m_CurrentCRC = CalculateCRCEnd(m_CurrentCRC);
				unsigned short int diskCRC = (m_headerData[5] << 8) + m_headerData[6];
				if(m_CurrentCRC != diskCRC){
					m_CRCError = true;
					m_WD1772State = wd1772_motoridle;
				}
				else{
					m_CRCError = false;
					m_WD1772State = wd1772_motoridle;
				};
				m_IntRQ = true;
				m_Busy = false;
				m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772DATATIMEOUT, MAXCYCLE);
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		//if mfm medium and double density selected
		if(m_trackIsValid && (m_encodingType >= 1) && (m_encodingType <= 3)  && m_IsDoubleDensity) {
			unsigned int tmpDataLong = GetTrackDataLongword(m_trackData, m_currentTrackOffset);
			tmpDataLong = tmpDataLong >> (16 - m_SyncBitShift);
			m_CurrentByte = m_headerData[m_WordsElapsed] = DecodeByteMFM(tmpDataLong & 0xffff);
			if(m_DataByteProvided)
				m_DataOverflow = true;
			m_DataByteProvided = true;
#ifdef TESTSTATE
			printf("Header data %2x at %2x\n", m_headerData[m_WordsElapsed], m_WordsElapsed);
#endif
			m_DRQ = true;
			++m_WordsElapsed;
			if(m_WordsElapsed == 7){
				m_CurrentCRC = 0x2db3;
				for(int loop = 0; loop < 5; ++loop) {
					m_CurrentCRC = CalculateCRC(m_CurrentCRC, m_headerData[loop]);
				};
				m_CurrentCRC = CalculateCRCEnd(m_CurrentCRC);
				unsigned short int diskCRC = (m_headerData[5] << 8) + m_headerData[6];
				if(m_CurrentCRC != diskCRC){
					m_CRCError = true;
					m_WD1772State = wd1772_motoridle;
				}
				else{
					m_CRCError = false;
					m_WD1772State = wd1772_motoridle;
					m_WordsElapsed = 0;
				};
				m_IntRQ = true;
				m_Busy = false;
				m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772DATATIMEOUT, MAXCYCLE);

			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		break;

	case wd1772_readuntilwritedata:
		//will skip 18 words
		//if fm and not double density selected
		if(m_trackIsValid && (m_encodingType == 0) && !m_IsDoubleDensity) {
			if(m_WordsElapsed == 2){
				m_DRQ = true;
			};
			if(m_WordsElapsed == 11){
				if(!m_DataByteProvided){
					m_IntRQ = true;
					m_Busy = false;
					m_DataOverflow = true;
					m_WD1772State = wd1772_motoridle;
				};
			};
			if(m_WordsElapsed == 18){
				m_WordsElapsed = 0;
				m_WD1772State = wd1772_writedata;
			}
			else{
				++m_WordsElapsed;
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		//if mfm medium and double density selected
		if(m_trackIsValid && (m_encodingType >= 1) && (m_encodingType <= 3) && m_IsDoubleDensity) {
			if(m_WordsElapsed == 2){
				m_DRQ = true;
			};
			if(m_WordsElapsed == 11){
				if(!m_DataByteProvided){
					m_IntRQ = true;
					m_Busy = false;
					m_DataOverflow = true;
					m_WD1772State = wd1772_motoridle;
				};
			};
			//11 from above, 1 gap +11 delay +12 zero-bytes in gap + 3 sync
			if(m_WordsElapsed == (11+24+3)){
				m_WordsElapsed = 0;
				m_WD1772State = wd1772_writedata;
			}
			else{
				++m_WordsElapsed;
			};
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		break;

	case wd1772_writedata:
		//if fm and not double density selected
		if(m_trackIsValid && (m_encodingType == 0) && !m_IsDoubleDensity) {
			if(m_WordsElapsed == 0){
				//init CRC
				m_CurrentCRC = 0xffff;
				m_CurrentCRC = CalculateCRC(m_CurrentCRC, 0xFB);
			};
			if((m_WordsElapsed >= 1) && (m_WordsElapsed <= m_SectorLength)){
				//store byte in sectortable
				if(!m_DataByteProvided){
					m_DataOverflow = true;
					m_CurrentByte = 0;
				};
				if((m_WordsElapsed - 1) < m_SectorLengthInTrack){
					unsigned char * tmpPtr = m_SectorDataOnWrite + m_WordsElapsed - 1;
					*tmpPtr = m_CurrentByte;
				};
				//request new byte
				m_DataByteProvided = false;
				if(m_WordsElapsed < m_SectorLength)
					m_DRQ = true;
				//calculate CRC
				m_CurrentCRC = CalculateCRC(m_CurrentCRC, m_CurrentByte);
			};
			if(m_WordsElapsed == (m_SectorLength + 1 + 2 + 1)){
				//end of writing
				m_Busy = false;
				m_IntRQ = true;
				m_WD1772State = wd1772_motoridle;

#ifdef TESTSTATE
	if(m_SectorDataOnWrite == &m_SectorWriteDummy[0]){
		printf("Dummy Write\n");
	}
	else{
		printf("Ordinary Write\n");
	};
#endif
				//write sector data into WATR buffer
				if(m_SectorDataOnWrite != &m_SectorWriteDummy[0]){
					unsigned char side = m_IsDiskSide1 ? 1 : 0;
					if((m_WATRHeader.SidePtr[0] == NULL) && (m_WATRHeader.SidePtr[1] == NULL)){
						m_WATRHeader.DiskSectorLen = m_SectorLengthInTrack;
					};
					if(m_WATRHeader.SidePtr[side] == NULL){
						m_WATRHeader.SidePtr[side] = new WATRSide;
						for(int zloop = 0; zloop < 40; ++zloop){
							m_WATRHeader.SidePtr[side]->TrackPtr[zloop] = NULL;
						};
					};
					if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack] == NULL){
						m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack] = new WATRTrack;
						for(int zloop = 0; zloop < 30; ++zloop){
							m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[zloop] = NULL;
						};
					};
					//now lets see if we can find the sector
					int iloop = 0;
					bool found = false;
					while((iloop < 30) && !found){
						if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop] != NULL){
							if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorNumber == (m_WantedSector - 1) && 
								(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorSequence == m_CurrentSectorSequenceNumber)){
								found = true;
								break;
							};
						};
						++iloop;
					};
					if(!found){
						//create all sector entries up to sectorsequencenumber
						for(int jloop = 0; jloop <= m_CurrentSectorSequenceNumber; ++jloop){
							bool found2 = false;
							int kloop = 0;
							while((kloop < 30) && !found2){
								if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop] != NULL){
									if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop]->SectorNumber == (m_WantedSector - 1) && 
										(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop]->SectorSequence == jloop)){
										found2 = true;
									};
								};
								++kloop;
							};
							if(!found2){
								kloop = 0;
								bool inserted = false;
								while((kloop < 30) && !inserted){
									if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop] == NULL){
										m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop] = new WATRSectors;
										m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop]->SectorNumber = m_WantedSector - 1;
										m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop]->SectorSequence = jloop;
										m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop]->SectorOffset = m_SectorLayout[m_WantedSector - 1][jloop].OffsetInProcessed;
										m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop]->SectorData = NULL;
										inserted = true;
										break;
									};
									++kloop;
								};
							};
						};
					};
					iloop = 0;
					found = false;
					while((iloop < 30) && !found){
						if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop] != NULL){
							if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorNumber == (m_WantedSector - 1) && 
								(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorSequence == m_CurrentSectorSequenceNumber)){
								found = true;
								break;
							};
						};
						++iloop;
					};
					if(found){
						if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorData == NULL){
							m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorData = new unsigned char[m_WATRHeader.DiskSectorLen];
						};
						memcpy(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorData, m_SectorDataOnWrite, m_WATRHeader.DiskSectorLen);
					};
					//save WATR
					SaveWATRData();
				};
			};

			++m_WordsElapsed;
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		//if mfm medium and double density selected
		if(m_trackIsValid && (m_encodingType >= 1) && (m_encodingType <= 3) && m_IsDoubleDensity) {
			if(m_WordsElapsed == 0){
				//init CRC
				m_CurrentCRC = 0x2db3;
				m_CurrentCRC = CalculateCRC(m_CurrentCRC, 0xFB);
			};
			if((m_WordsElapsed >= 1) && (m_WordsElapsed <= m_SectorLength)){
				//store byte in sectortable
				if(!m_DataByteProvided){
					m_DataOverflow = true;
					m_CurrentByte = 0;
				};
				if((m_WordsElapsed - 1) < m_SectorLengthInTrack){
					unsigned char * tmpPtr = m_SectorDataOnWrite + m_WordsElapsed - 1;
					*tmpPtr = m_CurrentByte;
				};
				//request new byte
				m_DataByteProvided = false;
				if(m_WordsElapsed < m_SectorLength)
					m_DRQ = true;
				//calculate CRC
				m_CurrentCRC = CalculateCRC(m_CurrentCRC, m_CurrentByte);
			};
			if(m_WordsElapsed == (m_SectorLength + 1 + 2 + 1)){
				//end of writing
				m_Busy = false;
				m_IntRQ = true;
				m_WD1772State = wd1772_motoridle;

				//write sector data into WATR buffer
				if(m_SectorDataOnWrite != &m_SectorWriteDummy[0]){
					unsigned char side = m_IsDiskSide1 ? 1 : 0;
					if((m_WATRHeader.SidePtr[0] == NULL) && (m_WATRHeader.SidePtr[1] == NULL)){
						m_WATRHeader.DiskSectorLen = m_SectorLengthInTrack;
					};
					if(m_WATRHeader.SidePtr[side] == NULL){
						m_WATRHeader.SidePtr[side] = new WATRSide;
						for(int zloop = 0; zloop < 40; ++zloop){
							m_WATRHeader.SidePtr[side]->TrackPtr[zloop] = NULL;
						};
					};
					if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack] == NULL){
						m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack] = new WATRTrack;
						for(int zloop = 0; zloop < 30; ++zloop){
							m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[zloop] = NULL;
						};
					};
					//now lets see if we can find the sector
					int iloop = 0;
					bool found = false;
					while((iloop < 30) && !found){
						if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop] != NULL){
							if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorNumber == (m_WantedSector - 1) && 
								(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorSequence == m_CurrentSectorSequenceNumber)){
								found = true;
								break;
							};
						};
						++iloop;
					};
					if(!found){
						//create all sector entries up to sectorsequencenumber
						for(int jloop = 0; jloop <= m_CurrentSectorSequenceNumber; ++jloop){
							bool found2 = false;
							int kloop = 0;
							while((kloop < 30) && !found2){
								if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop] != NULL){
									if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop]->SectorNumber == (m_WantedSector - 1) && 
										(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop]->SectorSequence == jloop)){
										found2 = true;
									};
								};
								++kloop;
							};
							if(!found2){
								kloop = 0;
								bool inserted = false;
								while((kloop < 30) && !inserted){
									if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop] == NULL){
										m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop] = new WATRSectors;
										m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop]->SectorNumber = m_WantedSector - 1;
										m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop]->SectorSequence = jloop;
										m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop]->SectorOffset = m_SectorLayout[m_WantedSector - 1][jloop].OffsetInProcessed;
										m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[kloop]->SectorData = NULL;
										inserted = true;
										break;
									};
									++kloop;
								};
							};
						};
					};
					iloop = 0;
					found = false;
					while((iloop < 30) && !found){
						if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop] != NULL){
							if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorNumber == (m_WantedSector - 1) && 
								(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorSequence == m_CurrentSectorSequenceNumber)){
								found = true;
								break;
							};
						};
						++iloop;
					};
					if(found){
						if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorData == NULL){
							m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorData = new unsigned char[m_WATRHeader.DiskSectorLen];
						};
						memcpy(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorData, m_SectorDataOnWrite, m_WATRHeader.DiskSectorLen);
					};
					//save WATR
					SaveWATRData();
				};
			};

			++m_WordsElapsed;
			CountTrackDataOffset();
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, CyclesToNextDataWord());
		};
		break;

	default:
		;
	};
//	(this->* m_DataFunc)();
};

void WD1772Real::CheckHeadStepDelay(){
	switch(m_WD1772State){
	case wd1772_restore:
		m_totalTime = (m_CPU8040->GetCPUCycle() - m_MotorStartCycle)/8333333;
		if (m_CurrentTrack) {	
			--m_CurrentTrack;
		};
		CreateTrackData();
		PrepareMotorOff();
		if(m_WantedTrack != m_CurrentTrack) {
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772HEADSTEP, 6*8333);
		}
		else {
			m_Busy = false;
			m_WD1772State = wd1772_motoridle;
			//DestTrackReached();
			m_IntRQ = true;
		};
//	if(m_StepIn) {
//		++m_CurrentTrack;
//	}
//	else {
//		if (m_CurrentTrack) {	
//			--m_CurrentTrack;
//		};
//	};
//	m_currentTrackByte = (tmpOldCurrentOffset * m_EncodedBytesPerTrack / tmpOldTrackLen) & 0xfffe;
		break;
	case wd1772_stepin:
		m_totalTime = (m_CPU8040->GetCPUCycle() - m_MotorStartCycle)/8333333;
		++m_CurrentTrack;
		CreateTrackData();
		PrepareMotorOff();
		m_Busy = false;
		m_WD1772State = wd1772_motoridle;
		m_IntRQ = true;
		break;

	case wd1772_seek:
		m_totalTime = (m_CPU8040->GetCPUCycle() - m_MotorStartCycle)/8333333;
		if(m_CurrentTrack < m_WantedTrack){
			++m_CurrentTrack;
		}
		else if(m_CurrentTrack > m_WantedTrack){
			--m_CurrentTrack;
		};
		CreateTrackData();
		if(m_CurrentTrack != m_WantedTrack){
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772HEADSTEP, 6*8333);
		}
		else{
			PrepareMotorOff();
			m_Busy = false;
			m_WD1772State = wd1772_motoridle;
			m_IntRQ = true;
		};
		break;

	default:
		;
	};
};

void WD1772Real::CheckHeadCalmDelay(){
};

void WD1772Real::CheckMotor(){
	(this->* m_MotorFunc)();
};

void WD1772Real::DataSearchTimeOut(){
	m_Busy = false;
	m_IntRQ = true;
	m_RecordNotFound = true;
	m_WD1772State = wd1772_motoridle;
};

void WD1772Real::MotorOn(){
	if(m_MotorIsOn == false){
		//get current raw trackdata
		CreateTrackData();

		m_totalTime = 0;
		m_MotorStartCycle = m_CPU8040->GetCPUCycle();
//		m_currentTrackByte = 0;
	};
	//set motor on
	m_MotorIsOn = true;

//	//trigger offset evaluation
//	m_DataFunc = &WD1772Real::CountTrackDataOffset;
//	m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, m_CPUCyclesToNextEncodedWordOfTrack);

	PrepareMotorOff();
};

void WD1772Real::MotorOff(){
	m_MotorIsOn = false;
	m_WD1772State = wd1772_idle;
};

void WD1772Real::PrepareMotorOff(){
	//set motor off after 2 seconds
	m_MotorFunc = &WD1772Real::MotorOff;
	m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772MOTOR, 2*8333333);
};

void WD1772Real::CountTrackDataOffset(){
	if(m_imageType == 1) {
		m_currentTrackOffset += 2;
		if(m_currentTrackOffset >= m_EncodedBytesPerTrack) {
			m_currentTrackOffset = 0;
		};
	};
#ifdef WITH_CAPS
	if((m_imageType == 2) && m_imageLoaded){
		m_currentTrackOffset += 2;
		if(m_currentTrackOffset >= m_EncodedBytesPerTrack) {
			m_CurrentRevolution++;
			m_CurrentRevolution %= 4;
			m_encodingType = AnalyzeFromCT();
			m_currentTrackOffset = 0;
		};
	};
#endif
};

unsigned char WD1772Real::ByteFromWD1772(){
#ifdef TESTSTATE
	printf("Byte from WD1772 Reg. %d\n", m_SelectedRegister);
#endif
	if(m_SelectedRegister == 0){
		return GetStatusRegister();
	};
	if(m_SelectedRegister == 1){
		return m_CurrentTrack;
	};
	if(m_SelectedRegister == 2){
		if(m_LastCommand == 0xcc){
			return m_headerData[1];
		};
		return m_WantedSector;
	};
	if(m_SelectedRegister == 3){
		m_DataByteProvided= false;
		return m_CurrentByte;
	};
	return 0;
};



void WD1772Real::ByteToWD1772(unsigned char myByte){
	//m_WD1772Byte = myByte;
#ifdef TESTSTATE
	printf("Byte to WD1772 Reg. %d data %2x\n", m_SelectedRegister, myByte);
#endif
	if(m_SelectedRegister == 0){
#ifdef TESTSTATE
		printf("WD1772Command: %2x\n", myByte);
#endif
		m_LastCommand = myByte;
		m_DRQ = false;
		//command: restore [seek track 0] (actually XF551: 0x08)
		if((myByte & 0xf0) == 0) {
			m_WD1772State = wd1772_restore;
//			m_StepIn = false;
			m_WantedTrack = 0;
//			if(myByte & 0x08) {
				if(!m_MotorIsOn){
					MotorOn();
				};
				if(m_CurrentTrack) {
					m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772HEADSTEP, 6*8333);
					m_Busy = true;
				}
				else {
					m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772HEADSTEP, 1);
				};
//			}
//			else {
//			};
			PrepareMotorOff();
		};
		//command: seek (actually XF551: 0x18)
		if((myByte & 0xf0) == 0x10) {
			m_WD1772State = wd1772_seek;
			m_WantedTrack = m_CurrentByte;
			m_Busy = true;
			if(!m_MotorIsOn){
				MotorOn();
			};
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772HEADSTEP, 6*8333);
		};
		//command: step-in (actually XF551: 0x48)
		if((myByte & 0xf0) == 0x40) {
			m_WD1772State = wd1772_stepin;
			m_WantedTrack = m_CurrentTrack + 1;
			m_Busy = true;
			if(!m_MotorIsOn){
				MotorOn();
			};
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772HEADSTEP, 6*8333);
		};
		//interrupt current command (actually XF551: 0xD0)
		if((myByte & 0xf0) == 0xd0) {
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772HEADSTEP, MAXCYCLE);
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772HEADCALM, MAXCYCLE);
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, MAXCYCLE);
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772DATATIMEOUT, MAXCYCLE);
			m_WD1772State = wd1772_motoridle;
			m_Busy = false;			
			PrepareMotorOff();
		};
		//read single sector (actually XF551: 0x88)
		if((myByte & 0xf0) == 0x80) {
			m_WD1772State = wd1772_readsearchheadersync;
			m_Busy = true;
			m_RecordNotFound = false;
			m_CRCError = false;
			m_DeletedDataMark = false;
			m_IntRQ = false;
			m_DataByteProvided = false;
			m_DataOverflow = false;
			//time out after 5 revolutions
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772DATATIMEOUT, 1*8333333);
			//determine offset of next dataword
			m_totalTime = (m_CPU8040->GetCPUCycle() - m_MotorStartCycle)/8333333.3;
			unsigned int tempRevolutions = m_totalTime / (m_Is288RPM ? 0.208333333 : 0.2);
#ifdef WITH_CAPS
			if((m_imageType == 2) && (tempRevolutions != m_OldRevolution) && m_imageLoaded){
				++m_CurrentRevolution;
				m_CurrentRevolution %= 4;
				m_encodingType = AnalyzeFromCT();
			};
			m_OldRevolution = tempRevolutions;
#endif
			double tempTimedOffset = m_totalTime - tempRevolutions * (m_Is288RPM ? 0.208333333 : 0.2);
#ifdef TESTSTATE3
//			printf("tempTimedOffset %f, m_OffsetStartTrack %f\n", tempTimedOffset, m_OffsetStartTrack);
#endif
			if(tempTimedOffset < m_OffsetStartTrack){
				tempTimedOffset += (m_Is288RPM ? 0.208333333 : 0.2);
			};
			m_currentTrackOffset = ((tempTimedOffset - m_OffsetStartTrack) / (m_Is288RPM ? 0.208333333 : 0.2 )* m_EncodedBytesPerTrack);
			m_currentTrackOffset &= 0xfffe;
			//time to next track data word
			int nexttime = CyclesToNextDataWord();
//			int nexttime = ((((m_currentTrackOffset + 2.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2)) - ((m_currentTrackOffset + 0.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2))) * 8333333.333);
#ifdef TESTSTATE
			printf("currentTrackByte: %d, Total: %d\n", m_currentTrackOffset, m_EncodedBytesPerTrack);
			printf("test: %f\n", (((m_currentTrackOffset + 2.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2)) - ((m_currentTrackOffset + 0.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2))) * 8333333.333);
			printf("nexttime: %d, m_Is288RPM:%d\n",nexttime, m_Is288RPM);
#endif

#ifdef TESTSTATE3
//			printf("wanted sector: %d, wanted track: %d, totaltime %f, CTOffset %4x\n", m_WantedSector, m_WantedTrack, m_totalTime, m_currentTrackOffset);
#endif
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, nexttime);
		};
		//read address (actually XF551: 0xcc)
		if((myByte & 0xf0) == 0xc0) {
			m_WD1772State = wd1772_rasearchheadersync;
			m_Busy = true;
			m_RecordNotFound = false;
			m_CRCError = false;
			m_IntRQ = false;
			m_DataByteProvided = false;
			m_DataOverflow = false;
			//time out after 5 revolutions
			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772DATATIMEOUT, 1*8333333);
			//determine offset of next dataword
			m_totalTime = (m_CPU8040->GetCPUCycle() - m_MotorStartCycle)/8333333.3;
			unsigned int tempRevolutions = m_totalTime / (m_Is288RPM ? 0.208333333 : 0.2);
#ifdef WITH_CAPS
			if((m_imageType == 2) && (tempRevolutions != m_OldRevolution) && m_imageLoaded){
				++m_CurrentRevolution;
				m_CurrentRevolution %= 4;
				m_encodingType = AnalyzeFromCT();
			};
			m_OldRevolution = tempRevolutions;
#endif
			double tempTimedOffset = m_totalTime - tempRevolutions * (m_Is288RPM ? 0.208333333 : 0.2);
			if(tempTimedOffset < m_OffsetStartTrack){
				tempTimedOffset += (m_Is288RPM ? 0.208333333 : 0.2);
			};
			m_currentTrackOffset = ((tempTimedOffset - m_OffsetStartTrack) / (m_Is288RPM ? 0.208333333 : 0.2) * m_EncodedBytesPerTrack);
			m_currentTrackOffset &= 0xfffe;
			//time to next track data word
			int nexttime = CyclesToNextDataWord();
//			int nexttime = ((((m_currentTrackOffset + 2.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2)) - ((m_currentTrackOffset + 0.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2))) * 8333333.333);
#ifdef TESTSTATE
			printf("currentTrackByte: %d, Total: %d\n", m_currentTrackOffset, m_EncodedBytesPerTrack);
			printf("test: %f\n", (((m_currentTrackOffset + 2.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2)) - ((m_currentTrackOffset + 0.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2))) * 8333333.333);
			printf("nexttime: %d, m_Is288RPM:%d\n",nexttime, m_Is288RPM);
#endif

			m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, nexttime);
#ifdef TESTSTATE
			printf("Valid Track on WD command CC: %d\n", m_trackIsValid);
#endif
		};
		//command: write (actually XF551: 0xa2)
		if((myByte & 0xf0) == 0xa0) {
			m_WD1772State = wd1772_readsearchheadersync;
			m_Busy = true;
			m_RecordNotFound = false;
			m_CRCError = false;
			m_DeletedDataMark = false;
			m_IntRQ = false;
			m_DataByteProvided = false;
			m_DataOverflow = false;
			if(m_WriteProtect){
				m_Busy = false;
				m_IntRQ = true;
			}
			else{
				//time out after 5 revolutions
				m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772DATATIMEOUT, 1*8333333);
				//determine offset of next dataword
				m_totalTime = (m_CPU8040->GetCPUCycle() - m_MotorStartCycle)/8333333.3;
				unsigned int tempRevolutions = m_totalTime / (m_Is288RPM ? 0.208333333 : 0.2);
#ifdef WITH_CAPS
				if((m_imageType == 2) && (tempRevolutions != m_OldRevolution) && m_imageLoaded){
					++m_CurrentRevolution;
					m_CurrentRevolution %= 4;
					m_encodingType = AnalyzeFromCT();
				};
				m_OldRevolution = tempRevolutions;
#endif
				double tempTimedOffset = m_totalTime - tempRevolutions * (m_Is288RPM ? 0.208333333 : 0.2);
#ifdef TESTSTATE3
//				printf("tempTimedOffset %f, m_OffsetStartTrack %f\n", tempTimedOffset, m_OffsetStartTrack);
#endif
				if(tempTimedOffset < m_OffsetStartTrack){
					tempTimedOffset += (m_Is288RPM ? 0.208333333 : 0.2);
				};
				m_currentTrackOffset = ((tempTimedOffset - m_OffsetStartTrack) / (m_Is288RPM ? 0.208333333 : 0.2 )* m_EncodedBytesPerTrack);
				m_currentTrackOffset &= 0xfffe;
				//time to next track data word
				int nexttime = CyclesToNextDataWord();
//				int nexttime = ((((m_currentTrackOffset + 2.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2)) - ((m_currentTrackOffset + 0.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2))) * 8333333.333);
#ifdef TESTSTATE
				printf("currentTrackByte: %d, Total: %d\n", m_currentTrackOffset, m_EncodedBytesPerTrack);
				printf("test: %f\n", (((m_currentTrackOffset + 2.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2)) - ((m_currentTrackOffset + 0.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2))) * 8333333.333);
				printf("nexttime: %d, m_Is288RPM:%d\n",nexttime, m_Is288RPM);
#endif

#ifdef TESTSTATE3
//				printf("wanted sector: %d, wanted track: %d, totaltime %f, CTOffset %4x\n", m_WantedSector, m_WantedTrack, m_totalTime, m_currentTrackOffset);
#endif
				m_CPU8040->SetNextSpecialEvent(CPUEVENTWD1772NEXTDATA, nexttime);
			};
		};
	};
	if(m_SelectedRegister == 1){
		m_WantedTrack = myByte;
	};
	if(m_SelectedRegister == 2){
		m_WantedSector = myByte;
	};
	if(m_SelectedRegister == 3){
		m_CurrentByte = myByte;
		m_DataByteProvided = true;
	};
};

unsigned char WD1772Real::GetStatusRegister(){
	unsigned char tmpStatus;
	//bit 7: motor status
	tmpStatus = m_MotorIsOn ? 0x80 : 0;
	//
	//bit 6: write protect on any write command, else 0
	if	((m_LastCommand & 0xf0) == 0xa0)
		tmpStatus += m_WriteProtect ? 0x40 : 0;
	//bit 5: type 1 command: spin up completed
	if ((m_LastCommand & 0x80) == 0) {
		tmpStatus += 0x20;
	}
	//else set on deleted data mark
	else {
		tmpStatus += m_DeletedDataMark ? 0x20 : 0;
	};
	//bit 4: record not found
	tmpStatus += m_RecordNotFound ? 0x10 : 0;
	//bit 3: crc error
	tmpStatus += m_CRCError ? 0x08 : 0;
	//bit 2: type 1 command: track 0 status
	if ((m_LastCommand & 0x80) == 0) {
		tmpStatus += m_CurrentTrack ? 0x4 : 0;
	}
	//else computer did not respond to DRQ in time
	else {
		tmpStatus += m_DataOverflow ? 0x4 : 0;
		//m_DataOverflow = false;
	};
	//bit 1: on type 1 commands, status of index pulse signal (here ignored)
	//bit 1: copy of DRQ output
	tmpStatus += m_DataByteProvided ? 0x2 : 0;
	//bit0: wd1772 is busy
	tmpStatus += m_Busy ? 1 : 0;
	return tmpStatus;
};

#ifdef WITH_CAPS
char WD1772Real::AnalyzeFromCT(){
	//check & process MFM first
	unsigned int dataOffset = 0;
	unsigned int dataOffsetGap = 0;
	signed char syncShift = -1;
	short int sectorSize;
	char sectorOccurred[26];
	for(int iloop = 0; iloop < 26; ++iloop){
		sectorOccurred[iloop] = 0;
	};
	unsigned char headerData[7];
	dataOffset = 2;
	dataOffsetGap = 2;
	unsigned long int trackDataWord;
	bool found = false;
	//gap detection, in the first part set gap after a complete sector
	while((dataOffset < 0x1100) && !found){
		dataOffset += 2;
		//get longword
		trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
		//try to find a header sync in the longword
		syncShift = SearchSync(trackDataWord, 0x4489, 0x4489); //generic sync
		if(syncShift != -1){
			//loop until syncs are exhausted
			while(true){
				trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
				trackDataWord = trackDataWord >> (16 - syncShift);
				if((trackDataWord & 0xffff) != 0x4489){
					break;
				};
				dataOffset += 2;
			};

			//get first byte after syncs and check for 0xFE (header)
			trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
			trackDataWord = trackDataWord >> (16 - syncShift);
			unsigned char tmpByte = DecodeByteMFM(trackDataWord & 0xffff);
			if(tmpByte != 0xFE)
				continue;

			//header found, read it and check checksum
			for(int iloop = 0; iloop < 7; ++iloop){
				trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
				trackDataWord = trackDataWord >> (16 - syncShift);
				headerData[iloop] = DecodeByteMFM(trackDataWord);
				dataOffset += 2;
			};
			unsigned short int currentCRC = 0x2db3;
			for(int loop = 0; loop < 5; ++loop) {
				currentCRC = CalculateCRC(currentCRC, headerData[loop]);
			};
			currentCRC = CalculateCRCEnd(currentCRC);
			unsigned short int diskCRC = (headerData[5] << 8) + headerData[6];
#ifdef TESTSTATE3
    printf("CRC header - disk: %2x, calc: %2x\n", diskCRC, currentCRC);
#endif

			//if header CRC matches, get databytes of sector
			if(currentCRC == diskCRC){
				//search data mark
				dataOffset -= 2;
				for(int iloop = 0; iloop < 43; ++iloop){
					dataOffset += 2;
					//get longword
					trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
					//try to find a header sync in the longword
					syncShift = SearchSync(trackDataWord, 0x4489, 0x4489); //generic sync
					if(syncShift != -1){
						break;
					};
				};
				if(syncShift != -1){
					//loop until syncs are exhausted
					while(true){
						trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
						trackDataWord = trackDataWord >> (16 - syncShift);
						if((trackDataWord & 0xffff) != 0x4489){
							break;
						};
						dataOffset += 2;
					};
					//get first byte after syncs and check for 0xFE (header)
					trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
					trackDataWord = trackDataWord >> (16 - syncShift);
					tmpByte = DecodeByteMFM(trackDataWord &0xffff);
					if((tmpByte != 0xFB) && (tmpByte != 0xF8))
						continue;

					unsigned char sectorData[1+2048+2];
					sectorSize = 0x80 << headerData[4];
					if(sectorSize > 2048)
						sectorSize = 2048;
					for(int iloop = 0; iloop < 1+sectorSize+2; ++iloop){
						trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
						trackDataWord = trackDataWord >> (16 - syncShift);
						sectorData[iloop] = DecodeByteMFM(trackDataWord);
						dataOffset += 2;
					};
					currentCRC = 0x2db3;
					for(int loop = 0; loop < 1+sectorSize; ++loop) {
						currentCRC = CalculateCRC(currentCRC, sectorData[loop]);
					};
					currentCRC = CalculateCRCEnd(currentCRC);
					diskCRC = (sectorData[1+sectorSize] << 8) + sectorData[2+sectorSize];
#ifdef TESTSTATE3
    printf("CRC sector - disk: %2x, calc: %2x\n", diskCRC, currentCRC);
#endif
					if(currentCRC == diskCRC){
						found = true;
						dataOffsetGap = dataOffset;
					};
				}; // end datasync found
			}; //end headerCRC ok
		}; //end headersync found
	};
	//gap detection, in the second part set gap before a detected sector
	while((dataOffset < (m_CAPSTrackInfo[m_CurrentRevolution].tracklen >> 3)) && !found){
		dataOffsetGap = dataOffset;
		dataOffset += 2;
		//get longword
		trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
		//try to find a header sync in the longword
		syncShift = SearchSync(trackDataWord, 0x4489, 0x4489); //generic sync
		if(syncShift != -1){
			//loop until syncs are exhausted
			while(true){
				trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
				trackDataWord = trackDataWord >> (16 - syncShift);
				if((trackDataWord & 0xffff) != 0x4489){
					break;
				};
				dataOffset += 2;
			};

			//get first byte after syncs and check for 0xFE (header)
			trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
			trackDataWord = trackDataWord >> (16 - syncShift);
			unsigned char tmpByte = DecodeByteMFM(trackDataWord & 0xffff);
			if(tmpByte != 0xFE)
				continue;

			//header found, read it and check checksum
			for(int iloop = 0; iloop < 7; ++iloop){
				trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
				trackDataWord = trackDataWord >> (16 - syncShift);
				headerData[iloop] = DecodeByteMFM(trackDataWord);
				dataOffset += 2;
			};
			unsigned short int currentCRC = 0x2db3;
			for(int loop = 0; loop < 5; ++loop) {
				currentCRC = CalculateCRC(currentCRC, headerData[loop]);
			};
			currentCRC = CalculateCRCEnd(currentCRC);
			unsigned short int diskCRC = (headerData[5] << 8) + headerData[6];
			//if header CRC matches, set found
#ifdef TESTSTATE3
    printf("p2: CRC sector - disk: %2x, calc: %2x\n", diskCRC, currentCRC);
#endif
			if(currentCRC == diskCRC){
						found = true;
#ifdef TESTSTATE3
    printf("found.\n", diskCRC, currentCRC);
#endif
			};
		}; // end headersync found
	};
	if(found){
#ifdef TESTSTATE3
	    printf("track is %d, dataOffsetGap is %4x, tracklen is %x\n", m_CurrentTrack, dataOffsetGap, 
			m_CAPSTrackInfo[m_CurrentRevolution].tracklen >> 3);
#endif

		//now that a gap has been found, copy the track
		bool finished = false;
		unsigned int dataOffsetWrite = 0;
		dataOffset = dataOffsetGap;
		dataOffset -= 2;
		unsigned short dataOffsetHeader = 0xc000;
		SetTrackDataWord(&m_trackData[dataOffsetWrite], 0x2AAA);
		while(!finished) {
			//read data from track and write it as MFM to internal data track
			dataOffset += 2;
			dataOffsetWrite += 2;
			//get longword
			trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
			//try to find any sync in the longword
			syncShift = SearchSync(trackDataWord, 0x4489, 0x4489); //generic header
			//if sync found
			if(syncShift != -1){
				//loop until syncs are exhausted
				while(true){
					trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
					trackDataWord = trackDataWord >> (16 - syncShift);
					if((trackDataWord & 0xffff) != 0x4489){
						break;
					};
					dataOffset += 2;
				};

				//get first byte after syncs
				trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
				trackDataWord = trackDataWord >> (16 - syncShift);
				unsigned char whichSync = DecodeByteMFM(trackDataWord & 0xffff);
				unsigned int bytesToCopy = 0;

				//if no valid mark, jump out
				if((whichSync != 0xFE) && (whichSync != 0xFB) && (whichSync != 0xF8)){
					SetTrackDataWord(&m_trackData[dataOffsetWrite], 0x2aaa);
					continue;
				};
				//copy the correct amount of data for header or sector data
				if(whichSync == 0xFE){
					//check sector header for correct crc for creating entry in writetable
					unsigned short int currentCRC = 0x2db3;
					for(int iloop = 0; iloop < 7; ++iloop){
						trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset+2*iloop);
						trackDataWord = trackDataWord >> (16 - syncShift);
						headerData[iloop] = DecodeByteMFM(trackDataWord);
						if(iloop < 5)
							currentCRC = CalculateCRC(currentCRC, headerData[iloop]);
					};
					currentCRC = CalculateCRCEnd(currentCRC);
					unsigned short int diskCRC = (headerData[5] << 8) + headerData[6];
					if(currentCRC == diskCRC){
						char tmpSector = headerData[3] - 1;
						if(tmpSector < 26){
							if(m_SectorLayout[tmpSector][sectorOccurred[tmpSector]].SectorSequenceNumber == 0xff){
								m_SectorLayout[tmpSector][sectorOccurred[tmpSector]].SectorSequenceNumber = sectorOccurred[tmpSector];
								sectorOccurred[tmpSector]++;
								m_SectorLayout[tmpSector][sectorOccurred[tmpSector]].OffsetInProcessed = dataOffsetWrite;
							};
						};
					};

					bytesToCopy = 7;
					//get length field from header
					sectorSize = 0x80 << headerData[4];
					if(sectorSize > 2048)
						sectorSize = 2048;
					dataOffsetHeader = dataOffset;
				} 
				else{
					if((dataOffset - dataOffsetHeader) > (43+7)*2){
						SetTrackDataWord(&m_trackData[dataOffsetWrite], 0x2aaa);
						continue;
					};
					bytesToCopy = 1+sectorSize+2;
				};
#ifdef TESTSTATE3
				trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset + 3*2);
				trackDataWord = trackDataWord >> (16 - syncShift);
				unsigned char sector = DecodeByteMFM(trackDataWord);
			    printf("Recode: Sync is %1x, sector is %1d, offset is %4x, dataOffsetHeader is %4x, dataOffsetWrite is %4x\n", whichSync, sector, dataOffset, dataOffsetHeader, dataOffsetWrite);
#endif
				SetTrackDataWord(&m_trackData[dataOffsetWrite], 0x4489);
				dataOffsetWrite += 2;
				SetTrackDataWord(&m_trackData[dataOffsetWrite], 0x4489);
				dataOffsetWrite += 2;
				SetTrackDataWord(&m_trackData[dataOffsetWrite], 0x4489);
				dataOffsetWrite += 2;

				dataOffset -= 2;
				dataOffsetWrite -= 2;
				for(int iloop = 0; iloop < bytesToCopy; ++iloop){
					dataOffset += 2;
					dataOffsetWrite += 2;
					trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset);
					trackDataWord = trackDataWord >> (16 - syncShift);
					SetTrackDataWord(&m_trackData[dataOffsetWrite], trackDataWord & 0xffff);
				};
				//after copying sector data, reset size
				if((whichSync == 0xFB) || (whichSync == 0xF8))
					sectorSize = 0x80;
			}
			else{
				SetTrackDataWord(&m_trackData[dataOffsetWrite], 0x2AAA);
			};
			//finished when Offset > dataOffsetGap - 20 and sync found
			if (dataOffset > ((m_CAPSTrackInfo[m_CurrentRevolution].tracklen >> 3) + dataOffsetGap - 20)){
#ifdef TESTSTATE3
//			    printf("dataOffset is %4x\n", dataOffset);
#endif
				//get longword
				trackDataWord = GetTrackDataLongWordCT(m_CurrentRevolution, dataOffset+2);
				//try to find any sync in the longword
				syncShift = SearchSync(trackDataWord, 0x4489, 0x4489); //generic header
				if((syncShift != -1) || 
					(dataOffset > ((m_CAPSTrackInfo[m_CurrentRevolution].tracklen >> 3) + dataOffsetGap))){
#ifdef TESTSTATE3
    printf("Finished.\n");
#endif
					finished = true;
				};
			};
		};
		SetTrackDataWord(&m_trackData[dataOffsetWrite+2], 0x2aaa);
		//set trackstart and tracklen
		m_OffsetStartTrack = (m_Is288RPM ? 0.208333333 : 0.2) * dataOffsetGap / (m_CAPSTrackInfo[m_CurrentRevolution].tracklen >> 3);
		m_EncodedBytesPerTrack = (dataOffsetWrite + 2) & 0xfffe;
#ifdef TESTSTATE3
    	printf("OffsetStartTrack = %f, dataOffset %d, Tracklen %d\n", 
			m_OffsetStartTrack, dataOffset, m_EncodedBytesPerTrack);
#endif
		m_trackIsValid = true;

		//encoding MFM
		return 1;
	};

/****************************/
	//then check & process FM
	dataOffset = 4;
	dataOffsetGap = 4;
	uint64_t trackQuadWord = 0;
	syncShift = -1;
	found = false;
	//gap detection, in the first part set gap after a complete sector
	while((dataOffset < 0x1100) && !found){
		dataOffset += 4;
		//get quadword
		trackQuadWord = GetTrackDataQuadWordCT(m_CurrentRevolution, dataOffset);
		//try to find a header sync in the quadword
		syncShift = SearchSyncsCTFM(trackQuadWord, 0x55111554, 0x55111554, 0x55111554); //FE
		if(syncShift != -1){
			//header found, read it and check checksum
			unsigned char headerData[7];
			for(int iloop = 0; iloop < 7; ++iloop){
				trackQuadWord = GetTrackDataQuadWordCT(m_CurrentRevolution, dataOffset);
				trackQuadWord = trackQuadWord >> (32 - syncShift);
				headerData[iloop] = DecodeByteFMCT(trackQuadWord);
				dataOffset += 4;
			};
			unsigned short int currentCRC = 0xffff;
			for(int loop = 0; loop < 5; ++loop) {
				currentCRC = CalculateCRC(currentCRC, headerData[loop]);
			};
			currentCRC = CalculateCRCEnd(currentCRC);
			unsigned short int diskCRC = (headerData[5] << 8) + headerData[6];
#ifdef TESTSTATE3
    printf("CRC header - disk: %2x, calc: %2x\n", diskCRC, currentCRC);
#endif

			//if header CRC matches, get databytes of sector
			if(currentCRC == diskCRC){
				//search data mark
				dataOffset -= 4;
				for(int iloop = 0; iloop < 30; ++iloop){
					dataOffset += 4;
					//get quadword
					trackQuadWord = GetTrackDataQuadWordCT(m_CurrentRevolution, dataOffset);
					//try to find a header sync in the quadword
					syncShift = SearchSyncsCTFM(trackQuadWord, 0x55111455, 0x55111455, 0x55111444); //FB, F8
					if(syncShift != -1){
						break;
					};
				};
				if(syncShift != -1){
					unsigned char sectorData[1+2048+2];
					sectorSize = 0x80 << headerData[4];
					if(sectorSize > 2048)
						sectorSize = 2048;
					for(int iloop = 0; iloop < 1+sectorSize+2; ++iloop){
						trackQuadWord = GetTrackDataQuadWordCT(m_CurrentRevolution, dataOffset);
						trackQuadWord = trackQuadWord >> (32 - syncShift);
						sectorData[iloop] = DecodeByteFMCT(trackQuadWord);
						dataOffset += 4;
					};
					currentCRC = 0xffff;
					for(int loop = 0; loop < 1+sectorSize; ++loop) {
						currentCRC = CalculateCRC(currentCRC, sectorData[loop]);
					};
					currentCRC = CalculateCRCEnd(currentCRC);
					diskCRC = (sectorData[1+sectorSize] << 8) + sectorData[2+sectorSize];
#ifdef TESTSTATE3
    printf("CRC sector - disk: %2x, calc: %2x\n", diskCRC, currentCRC);
#endif
					if(currentCRC == diskCRC){
						found = true;
						dataOffsetGap = dataOffset;
					};
				}; // end datasync found
			}; //end headerCRC ok
		}; //end headersync found
	};
	//gap detection, in the second part set gap before a detected sector
	while((dataOffset < (m_CAPSTrackInfo[m_CurrentRevolution].tracklen >> 3)) && !found){
		dataOffsetGap = dataOffset;
		dataOffset += 4;
		//get quadword
		trackQuadWord = GetTrackDataQuadWordCT(m_CurrentRevolution, dataOffset);
		//try to find a header sync in the quadword
		syncShift = SearchSyncsCTFM(trackQuadWord, 0x55111554, 0x55111554, 0x55111554); //FE
		if(syncShift != -1){
			//header found, read it and check checksum
			for(int iloop = 0; iloop < 7; ++iloop){
				trackQuadWord = GetTrackDataQuadWordCT(m_CurrentRevolution, dataOffset);
				trackQuadWord = trackQuadWord >> (32 - syncShift);
				headerData[iloop] = DecodeByteFMCT(trackQuadWord);
				dataOffset += 4;
			};
			unsigned short int currentCRC = 0xffff;
			for(int loop = 0; loop < 5; ++loop) {
				currentCRC = CalculateCRC(currentCRC, headerData[loop]);
			};
			currentCRC = CalculateCRCEnd(currentCRC);
			unsigned short int diskCRC = (headerData[5] << 8) + headerData[6];
			//if header CRC matches, set found
#ifdef TESTSTATE3
    printf("p2: CRC sector - disk: %2x, calc: %2x\n", diskCRC, currentCRC);
#endif
			if(currentCRC == diskCRC){
						found = true;
#ifdef TESTSTATE3
    printf("found.\n", diskCRC, currentCRC);
#endif
			};
		}; // end headersync found
	};
	if(!found)
		dataOffsetGap = 4;
#ifdef TESTSTATE3
    printf("track is %d, dataOffsetGap is %4x, tracklen is %x\n", m_CurrentTrack, dataOffsetGap, 
		m_CAPSTrackInfo[m_CurrentRevolution].tracklen >> 3);
#endif

	//now that a gap has been found, recode the track from pseudo-mfm to fm
	bool finished = false;
	unsigned int dataOffsetWrite = 0;
	dataOffset = dataOffsetGap;
	dataOffset -= 4;
	unsigned short dataOffsetHeader = 0xc000;
	SetTrackDataWord(&m_trackData[dataOffsetWrite], 0xCCCC);
	while(!finished) {
		//read data from track and write it as FM to internal data track
		dataOffset += 4;
		dataOffsetWrite += 2;
		//get quadword
		trackQuadWord = GetTrackDataQuadWordCT(m_CurrentRevolution, dataOffset);
		//try to find any sync in the quadword
		syncShift = SearchSyncsCTFM(trackQuadWord, 0x55111554, 0x55111455, 0x55111444); //FE, FB, F8
		//if sync found
		if(syncShift != -1){
			// detect which sync
			trackQuadWord = trackQuadWord >> (32 - syncShift);
			unsigned char whichSync = DecodeByteFMCT(trackQuadWord);
			unsigned int bytesToCopy = 0;
			//copy the correct amount of data for header or sector data
			if(whichSync == 0xFE){
				//check sector header for correct crc for creating entry in writetable
				unsigned short int currentCRC = 0xffff;
				for(int iloop = 0; iloop < 7; ++iloop){
					trackQuadWord = GetTrackDataQuadWordCT(m_CurrentRevolution, dataOffset+4*iloop);
					trackQuadWord = trackQuadWord >> (32 - syncShift);
					headerData[iloop] = DecodeByteFMCT(trackQuadWord);
					if(iloop < 5)
						currentCRC = CalculateCRC(currentCRC, headerData[iloop]);
				};
				currentCRC = CalculateCRCEnd(currentCRC);
				unsigned short int diskCRC = (headerData[5] << 8) + headerData[6];
				if(currentCRC == diskCRC){
					char tmpSector = headerData[3] - 1;
					if(tmpSector < 26){
						if(m_SectorLayout[tmpSector][sectorOccurred[tmpSector]].SectorSequenceNumber == 0xff){
							m_SectorLayout[tmpSector][sectorOccurred[tmpSector]].SectorSequenceNumber = sectorOccurred[tmpSector];
							sectorOccurred[tmpSector]++;
							m_SectorLayout[tmpSector][sectorOccurred[tmpSector]].OffsetInProcessed = dataOffsetWrite;
						};
					};
				};

				bytesToCopy = 7;
				//get length field from header
				sectorSize = 0x80 << headerData[4];
				if(sectorSize > 2048)
					sectorSize = 2048;
				dataOffsetHeader = dataOffset;
			} 
			else{
				if((dataOffset - dataOffsetHeader) > (30+7)*4)
					continue;
				bytesToCopy = 1+sectorSize+2;
			};
#ifdef TESTSTATE3
			trackQuadWord = GetTrackDataQuadWordCT(m_CurrentRevolution, dataOffset + 3*4);
			trackQuadWord = trackQuadWord >> (32 - syncShift);
	unsigned char sector = DecodeByteFMCT(trackQuadWord);
    printf("Recode: Sync is %1x, sector is %1d, offset is %4x, dataOffsetHeader is %4x, dataOffsetWrite is %4x\n", whichSync, sector, dataOffset, dataOffsetHeader, dataOffsetWrite);
#endif
			dataOffset -= 4;
			dataOffsetWrite -= 2;
			unsigned int tmpDataRecoded = 0xCCCC;
			for(int iloop = 0; iloop < bytesToCopy; ++iloop){
				dataOffset += 4;
				dataOffsetWrite += 2;
				trackQuadWord = GetTrackDataQuadWordCT(m_CurrentRevolution, dataOffset);
				trackQuadWord = trackQuadWord >> (32 - syncShift);
				tmpDataRecoded = RecodeDataFMCT(trackQuadWord & 0xffffffff);
				SetTrackDataWord(&m_trackData[dataOffsetWrite], tmpDataRecoded);
			};
			//after copying sector data, reset size
			if((whichSync == 0xFB) || (whichSync == 0xF8))
				sectorSize = 0x80;
		}
		else{
			SetTrackDataWord(&m_trackData[dataOffsetWrite], 0xCCCC);
		};
		//finished when 
		// Offset > dataOffsetGap - 20 and header found
		if (dataOffset > ((m_CAPSTrackInfo[m_CurrentRevolution].tracklen >> 3) + dataOffsetGap - 20)){
#ifdef TESTSTATE3
//    printf("dataOffset is %4x\n", dataOffset);
#endif
			//get quadword
			trackQuadWord = GetTrackDataQuadWordCT(m_CurrentRevolution, dataOffset+4);
			//try to find any sync in the quadword
			syncShift = SearchSyncsCTFM(trackQuadWord, 0x55111554, 0x55111554, 0x55111554); //FE
			if((syncShift != -1) || 
				(dataOffset > ((m_CAPSTrackInfo[m_CurrentRevolution].tracklen >> 3) + dataOffsetGap))){
#ifdef TESTSTATE3
    printf("Finished.\n");
#endif
				finished = true;
			};
		};
	};
	//set trackstart and tracklen
	m_OffsetStartTrack = (m_Is288RPM ? 0.208333333 : 0.2) * dataOffsetGap / (m_CAPSTrackInfo[m_CurrentRevolution].tracklen >> 3);
	m_EncodedBytesPerTrack = (dataOffsetWrite + 2) & 0xfffe;
#ifdef TESTSTATE3
    printf("OffsetStartTrack = %f, dataOffset %d, Tracklen %d\n", 
		m_OffsetStartTrack, dataOffset, m_EncodedBytesPerTrack);
#endif
	m_trackIsValid = true;

	//encoding FM
	return 0;
};
#endif

void WD1772Real::CreateTrackData(){
	if(m_imageType == 1) {
		CreateTrackFromATR();
	};
#ifdef WITH_CAPS
	if((m_imageType == 2) && m_imageLoaded){
		for(int iloop = 0; iloop < 5; ++iloop){
			if(CAPSLockTrack(&m_CAPSTrackInfo[iloop], m_CAPSImgId, m_CurrentTrack*2, 
					(m_IsDiskSide1) && (m_CAPSMaxHead > m_CAPSMinHead) ? m_CAPSMinHead+1 : m_CAPSMinHead,
					DI_LOCK_INDEX|DI_LOCK_NOISE|DI_LOCK_TRKBIT|DI_LOCK_UPDATEFD) != imgeOk) throw;
		};
		++m_CurrentRevolution;
		m_CurrentRevolution %= 4;
		ResetTrackWriteProperties(false);
		//load trackData from internal WATR buffer
		TransferTrackWriteProperties();

		m_encodingType = AnalyzeFromCT();
	};
#endif
};

void WD1772Real::CreateTrackFromATR(){
#ifdef TESTSTATE
	printf("CreateTrackFromATR entered\n");
#endif
	ResetTrackWriteProperties(false);
	//load trackData from internal WATR buffer
	TransferTrackWriteProperties();

	// FM (Single) encoding
	if(m_encodingType == 0){
#ifdef TESTSTATE
		printf("CreateTrackFromATR FM detected\n");
#endif
		//fm
		m_SectorLengthInTrack = 0x80;
		unsigned char sectorsPerTrack = 18;
		unsigned char sectorOrder[] = {
			0, 2, 4, 6, 8, 10, 12, 14, 16,
			1, 3, 5, 7, 9, 11, 13, 15, 17};
		unsigned int trackOffset = 0;
		//pre-Gap
		SetTrackDataWord(&m_trackData[trackOffset], 0xaaaa);
		trackOffset += 2;
		SetTrackDataWord(&m_trackData[trackOffset], 0xaaaa);
		trackOffset += 1;
		//sectors
		for(int loop = 0; loop < sectorsPerTrack; ++loop){
			//sectorheader
			unsigned char headerData[7];
			headerData[0] = 0xfe;
			headerData[1] = m_CurrentTrack;
			headerData[2] = 0; //side
			//sector number starts with 1
			headerData[3] = sectorOrder[loop] + 1;
			headerData[4] = 0; //sector size: 2^(n+7) -> n = 0
			unsigned int hdrCRC = 0xffff;
			for(int loop2 = 0; loop2 < 5; ++loop2){
				hdrCRC = CalculateCRC(hdrCRC, headerData[loop2]);
			};
			hdrCRC = CalculateCRCEnd(hdrCRC);
			headerData[5] = (hdrCRC >> 8) & 0xff;
			headerData[6] = hdrCRC & 0xff;

			if(m_SectorLayout[sectorOrder[loop]][0].SectorSequenceNumber == 0xff){
				m_SectorLayout[sectorOrder[loop]][0].SectorSequenceNumber = 0;
				m_SectorLayout[sectorOrder[loop]][0].OffsetInProcessed = trackOffset & 0xfffe;
			};

			//0xfe, encoded especially
			SetTrackDataWord(&m_trackData[trackOffset], 0xf57e);
			trackOffset += 2;
			for (int loop2 = 1; loop2 < 7; ++loop2){
				SetTrackDataWord(&m_trackData[trackOffset], EncodeByteFM(headerData[loop2]));
				trackOffset += 2;
			};
			//headergap
			for (int loop2 = 0; loop2 < 17; ++loop2){
				SetTrackDataWord(&m_trackData[trackOffset], EncodeByteFM(0x00));
				trackOffset += 2;
			};
			//encode sector
			unsigned char * tmpATRData = &m_ATRData[0x10 + m_CurrentTrack*18*0x80 + sectorOrder[loop]*0x80];
//printf("Offset T %2x S %2x is %4x\n", m_CurrentTrack, sectorOrder[loop], m_CurrentTrack*18*0x80 + sectorOrder[loop]*0x80);
			unsigned int sectorCRC = 0xffff;
			sectorCRC = CalculateCRC(sectorCRC, 0xfb);
			//0xfb, encoded especially
			SetTrackDataWord(&m_trackData[trackOffset], 0xf56f);
			trackOffset += 2;
			for (int loop2 = 0; loop2 < 0x80; ++loop2) {
				//all data bytes are eored with 0xff on disk
				sectorCRC = CalculateCRC(sectorCRC, tmpATRData[loop2] ^ 0xff);
				SetTrackDataWord(&m_trackData[trackOffset], EncodeByteFM(tmpATRData[loop2] ^ 0xff));
//printf("assembled track data is %2x from %2x\n", tmpATRData[loop2], loop2);
				trackOffset += 2;
			};
			sectorCRC = CalculateCRCEnd(sectorCRC);
			SetTrackDataWord(&m_trackData[trackOffset], EncodeByteFM((sectorCRC >> 8) & 0xff));
			trackOffset += 2;
			SetTrackDataWord(&m_trackData[trackOffset], EncodeByteFM(sectorCRC & 0xff));
			trackOffset += 2;
			//intersectorgap
			for (int loop2 = 0; loop2 < 20; ++loop2){
				SetTrackDataWord(&m_trackData[trackOffset], EncodeByteFM(0x00));
				trackOffset += 2;
			};
		};
		//track-gap
		for (int loop2 = 0; loop2 < 20; ++loop2){
			SetTrackDataWord(&m_trackData[trackOffset], EncodeByteFM(0x00));
			trackOffset += 2;
		};
		SetTrackDataWord(&m_trackData[trackOffset], 0xaaaa);
		trackOffset += 1;
		m_EncodedBytesPerTrack = trackOffset;
#ifdef TESTSTATE
		printf("m_EncodedBytesPerTrack is %4x\n", m_EncodedBytesPerTrack);
#endif
		m_OffsetStartTrack = 0.1;
//		m_CPUCyclesToNextEncodedWordOfTrack = (m_Is288RPM ? 0.208333333 : 0.2) * 8333333.333 /m_EncodedBytesPerTrack * 2;
		m_trackIsValid = true;
		return;
	};

	//MFM (Medium) encoding
	if(m_encodingType == 1){
#ifdef TESTSTATE
		printf("CreateTrackFromATR MFM Medium detected\n");
#endif
		//mfm
		m_SectorLengthInTrack = 0x80;
		unsigned char sectorsPerTrack = 26;
		unsigned char sectorOrder[] = {
			0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24,
			1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23, 25};
		unsigned int trackOffset = 0;
		//pre-Gap
		SetTrackDataWord(&m_trackData[trackOffset], 0x2aaa);
		trackOffset += 2;
		SetTrackDataWord(&m_trackData[trackOffset], 0xaaaa);
		trackOffset += 1;
		//sectors
		for(int loop = 0; loop < sectorsPerTrack; ++loop){
			//sectorheader
			unsigned char headerData[7];
			headerData[0] = 0xfe;
			headerData[1] = m_CurrentTrack;
			headerData[2] = 0; //side
			//sector number starts with 1
			headerData[3] = sectorOrder[loop] + 1;
			headerData[4] = 0; //sector size: 2^(n+7) -> n = 0
			unsigned int hdrCRC = 0x2db3;
			for(int loop2 = 0; loop2 < 5; ++loop2){
				hdrCRC = CalculateCRC(hdrCRC, headerData[loop2]);
			};
			hdrCRC = CalculateCRCEnd(hdrCRC);
			headerData[5] = (hdrCRC >> 8) & 0xff;
			headerData[6] = hdrCRC & 0xff;

			if(m_SectorLayout[sectorOrder[loop]][0].SectorSequenceNumber == 0xff){
				m_SectorLayout[sectorOrder[loop]][0].SectorSequenceNumber = 0;
				m_SectorLayout[sectorOrder[loop]][0].OffsetInProcessed = trackOffset & 0xfffe;
			};

			//sync 3x
			SetTrackDataWord(&m_trackData[trackOffset], 0x4489);
			trackOffset += 2;
			SetTrackDataWord(&m_trackData[trackOffset], 0x4489);
			trackOffset += 2;
			SetTrackDataWord(&m_trackData[trackOffset], 0x4489);
			trackOffset += 2;
			for (int loop2 = 0; loop2 < 7; ++loop2){
				SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(headerData[loop2], m_trackData[trackOffset - 2] & 0x01));
				trackOffset += 2;
			};
			//headergap
			for (int loop2 = 0; loop2 < 34; ++loop2){
				SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(0x00, m_trackData[trackOffset - 2] & 0x01));
				trackOffset += 2;
			};
			//encode sector
			unsigned char * tmpATRData = &m_ATRData[0x10 + m_CurrentTrack*26*0x80 + sectorOrder[loop]*0x80];
//printf("Offset T %2x S %2x is %4x\n", m_CurrentTrack, sectorOrder[loop], m_CurrentTrack*18*0x80 + sectorOrder[loop]*0x80);
			SetTrackDataWord(&m_trackData[trackOffset], 0x4489);
			trackOffset += 2;
			SetTrackDataWord(&m_trackData[trackOffset], 0x4489);
			trackOffset += 2;
			SetTrackDataWord(&m_trackData[trackOffset], 0x4489);
			trackOffset += 2;
			unsigned int sectorCRC = 0x2db3;
			sectorCRC = CalculateCRC(sectorCRC, 0xfb);
			SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(0xfb, m_trackData[trackOffset - 2] & 0x01));
			trackOffset += 2;
			for (int loop2 = 0; loop2 < 0x80; ++loop2) {
				//all data bytes are eored with 0xff on disk
				sectorCRC = CalculateCRC(sectorCRC, tmpATRData[loop2] ^ 0xff);
				SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(tmpATRData[loop2] ^ 0xff, m_trackData[trackOffset - 2] & 0x01));
//printf("assembled track data is %2x from %2x\n", tmpATRData[loop2], loop2);
				trackOffset += 2;
			};
			sectorCRC = CalculateCRCEnd(sectorCRC);
			SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM((sectorCRC >> 8) & 0xff, m_trackData[trackOffset - 2] & 0x01));
			trackOffset += 2;
			SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(sectorCRC & 0xff, m_trackData[trackOffset - 2] & 0x01));
			trackOffset += 2;
			//intersectorgap
			for (int loop2 = 0; loop2 < 47; ++loop2){
				SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(0x00, m_trackData[trackOffset - 2] & 0x01));
				trackOffset += 2;
			};
		};
		//track-gap
		for (int loop2 = 0; loop2 < 224; ++loop2){
			SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(0x00, m_trackData[trackOffset - 2] & 0x01));
			trackOffset += 2;
		};
		SetTrackDataWord(&m_trackData[trackOffset], 0xaaaa);
		trackOffset += 1;
		m_EncodedBytesPerTrack = trackOffset;
#ifdef TESTSTATE
		printf("m_EncodedBytesPerTrack is %4x\n", m_EncodedBytesPerTrack);
#endif
		m_OffsetStartTrack = 0.1;
//		m_CPUCyclesToNextEncodedWordOfTrack = (m_Is288RPM ? 0.208333333 : 0.2) * 8333333.333 /m_EncodedBytesPerTrack * 2;
		m_trackIsValid = true;
		return;
	};

	//MFM (Double) encoding
	if((m_encodingType == 2) || (m_encodingType == 3)){
#ifdef TESTSTATE
		printf("CreateTrackFromATR MFM Double/Quad detected\n");
#endif
		//no backside for double
		if((m_encodingType == 2) && m_IsDiskSide1) {
			m_trackIsValid = false;
			return;
		};

		//mfm
		m_SectorLengthInTrack = 0x100;
		unsigned char sectorsPerTrack = 18;
		unsigned char sectorOrder[] = {
			0, 2, 4, 6, 8, 10, 12, 14, 16,
			1, 3, 5, 7, 9, 11, 13, 15, 17};
		unsigned int trackOffset = 0;
		//pre-Gap
		SetTrackDataWord(&m_trackData[trackOffset], 0x2aaa);
		trackOffset += 2;
		SetTrackDataWord(&m_trackData[trackOffset], 0xaaaa);
		trackOffset += 1;
		//sectors
		for(int loop = 0; loop < sectorsPerTrack; ++loop){
			//sectorheader
			unsigned char headerData[7];
			headerData[0] = 0xfe;
			headerData[1] = m_CurrentTrack;
			headerData[2] = m_IsDiskSide1 ? 1 : 0; //side
			//sector number starts with 1
			headerData[3] = sectorOrder[loop] + 1;
			headerData[4] = 1; //sector size: 2^(n+7) -> n = 1
			unsigned int hdrCRC = 0x2db3;
			for(int loop2 = 0; loop2 < 5; ++loop2){
				hdrCRC = CalculateCRC(hdrCRC, headerData[loop2]);
			};
			hdrCRC = CalculateCRCEnd(hdrCRC);
			headerData[5] = (hdrCRC >> 8) & 0xff;
			headerData[6] = hdrCRC & 0xff;

			if(m_SectorLayout[sectorOrder[loop]][0].SectorSequenceNumber == 0xff){
				m_SectorLayout[sectorOrder[loop]][0].SectorSequenceNumber = 0;
				m_SectorLayout[sectorOrder[loop]][0].OffsetInProcessed = trackOffset & 0xfffe;
			};

			//sync 3x
			SetTrackDataWord(&m_trackData[trackOffset], 0x4489);
			trackOffset += 2;
			SetTrackDataWord(&m_trackData[trackOffset], 0x4489);
			trackOffset += 2;
			SetTrackDataWord(&m_trackData[trackOffset], 0x4489);
			trackOffset += 2;
			for (int loop2 = 0; loop2 < 7; ++loop2){
				SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(headerData[loop2], m_trackData[trackOffset - 2] & 0x01));
				trackOffset += 2;
			};
			//headergap
			for (int loop2 = 0; loop2 < 34; ++loop2){
				SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(0x00, m_trackData[trackOffset - 2] & 0x01));
				trackOffset += 2;
			};
			unsigned char * tmpATRData;
			//encode sector, first 3 sectors on track 0 are only #$80 in ATR
			if(!m_IsDiskSide1 && (sectorOrder[loop] <= 2) && (m_CurrentTrack == 0)){
				tmpATRData = &m_ATRData[0x10 + m_CurrentTrack*18*0x80 + sectorOrder[loop]*0x80];
			}
			else if(!m_IsDiskSide1) {
				tmpATRData = &m_ATRData[0x10 - 0x180 + m_CurrentTrack*18*0x100 + sectorOrder[loop]*0x100];
//printf("Offset T %2x S %2x is %4x\n", m_CurrentTrack, sectorOrder[loop], m_CurrentTrack*18*0x80 + sectorOrder[loop]*0x80);
			}
			else {
				//back side backwards from track 39 to track 0, sectors also backwards from 18 to 1
				tmpATRData = &m_ATRData[0x10 - 0x180 + (0x2d0 * 0x100) + (39 - m_CurrentTrack)*18*0x100 + (17-sectorOrder[loop])*0x100];
			};
			SetTrackDataWord(&m_trackData[trackOffset], 0x4489);
			trackOffset += 2;
			SetTrackDataWord(&m_trackData[trackOffset], 0x4489);
			trackOffset += 2;
			SetTrackDataWord(&m_trackData[trackOffset], 0x4489);
			trackOffset += 2;
			unsigned int sectorCRC = 0x2db3;
			sectorCRC = CalculateCRC(sectorCRC, 0xfb);
			SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(0xfb, m_trackData[trackOffset - 2] & 0x01));
			trackOffset += 2;
			for (int loop2 = 0; loop2 < 0x100; ++loop2) {
				unsigned char tmpByte;
				if(!m_IsDiskSide1 && (sectorOrder[loop] <= 2) && (m_CurrentTrack == 0)) {
					if(loop2 < 0x80) {
						tmpByte = tmpATRData[loop2];
					}
					else {
						tmpByte = 0;
					};
				}
				else {
					tmpByte = tmpATRData[loop2];
				};

				//all data bytes are eored with 0xff on disk
				sectorCRC = CalculateCRC(sectorCRC, tmpByte ^ 0xff);
				SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(tmpByte ^ 0xff, m_trackData[trackOffset - 2] & 0x01));
//printf("assembled track data is %2x from %2x\n", tmpATRData[loop2], loop2);
				trackOffset += 2;
			};
			sectorCRC = CalculateCRCEnd(sectorCRC);
			SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM((sectorCRC >> 8) & 0xff, m_trackData[trackOffset - 2] & 0x01));
			trackOffset += 2;
			SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(sectorCRC & 0xff, m_trackData[trackOffset - 2] & 0x01));
			trackOffset += 2;
			//intersectorgap
			for (int loop2 = 0; loop2 < 27; ++loop2){
				SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(0x00, m_trackData[trackOffset - 2] & 0x01));
				trackOffset += 2;
			};
		};
		//track-gap
		for (int loop2 = 0; loop2 < 452; ++loop2){
			SetTrackDataWord(&m_trackData[trackOffset], EncodeByteMFM(0x00, m_trackData[trackOffset - 2] & 0x01));
			trackOffset += 2;
		};
		SetTrackDataWord(&m_trackData[trackOffset], 0xaaaa);
		trackOffset += 1;
		m_EncodedBytesPerTrack = trackOffset;
#ifdef TESTSTATE
		printf("m_EncodedBytesPerTrack is %4x\n", m_EncodedBytesPerTrack);
#endif
		m_OffsetStartTrack = 0.1;
//		m_CPUCyclesToNextEncodedWordOfTrack = (m_Is288RPM ? 0.208333333 : 0.2) * 8333333.333 /m_EncodedBytesPerTrack * 2;
		m_trackIsValid = true;
		return;
	};
	m_trackIsValid = false;
};

unsigned int WD1772Real::EncodeByteFM(unsigned char byteToEncode){
	unsigned char tmpByte = byteToEncode;
	unsigned int tmpWord = 0;
	for (int loop = 0; loop < 8; ++loop){
		tmpWord = tmpWord << 1;
		tmpWord += 1;
		tmpWord = tmpWord << 1;
		tmpWord += (tmpByte & 0x80) ? 1 : 0;
		tmpByte = tmpByte << 1;
	};
//printf("encoded %2x as %4x\n", byteToEncode, tmpWord);
	return tmpWord;
};

unsigned char WD1772Real::DecodeByteFM(unsigned int byteToDecode){
	unsigned int tmpWord = byteToDecode;
	unsigned char tmpByte = 0;
	for (int loop = 0; loop < 8; ++loop){
		tmpByte = tmpByte << 1;
		tmpByte += (tmpWord & 0x4000) ? 1 : 0;
		tmpWord = tmpWord << 2;
	};
//printf("Decoded %4x as %2x\n", byteToDecode, tmpByte);
	return tmpByte;
};

#ifdef WITH_CAPS
unsigned char WD1772Real::DecodeByteFMCT(unsigned long byteToDecode){
	unsigned long tmpDWord = byteToDecode;
	unsigned char tmpByte = 0;
	for (int loop = 0; loop < 8; ++loop){
		tmpByte = tmpByte << 1;
		tmpByte += (tmpDWord & 0x10000000) ? 1 : 0;
		tmpDWord = tmpDWord << 4;
	};
//printf("Decoded %4x as %2x\n", byteToDecode, tmpByte);
	return tmpByte;
};

unsigned int WD1772Real::RecodeDataFMCT(unsigned long recodeLongWord){
	unsigned long tmpDataIn = recodeLongWord & 0x55555555;
	unsigned int tmpDataOut = 0;

	for(int iloop = 0; iloop < 8; ++iloop){
		tmpDataOut <<=2;
		tmpDataOut |= (tmpDataIn & 0x40000000) ? 2 : 0;
		tmpDataOut |= (tmpDataIn & 0x10000000) ? 1 : 0;
		tmpDataIn <<= 4;
	};
	return tmpDataOut;
};
#endif

unsigned int WD1772Real::EncodeByteMFM(unsigned char byteToEncode, bool previousBit){
	unsigned char tmpByte = byteToEncode;
	unsigned int tmpWord = previousBit ? 1 : 0;
	for (int loop = 0; loop < 8; ++loop){
		tmpWord = tmpWord << 2;
		tmpWord += (tmpByte & 0x80) ? 1 : 0;
		tmpWord += (tmpWord & 0x05) ? 0 : 2;
		tmpByte = tmpByte << 1;
	};
	return tmpWord;
};

unsigned char WD1772Real::DecodeByteMFM(unsigned int byteToDecode){
	unsigned int tmpWord = byteToDecode;
	unsigned char tmpByte = 0;
	for (int loop = 0; loop < 8; ++loop){
		tmpByte = tmpByte << 1;
		tmpByte += (tmpWord & 0x4000) ? 1 : 0;
		tmpWord = tmpWord << 2;
	};
//printf("Decoded %4x as %2x\n", byteToDecode, tmpByte);
	return tmpByte;
};

unsigned int WD1772Real::CyclesToNextDataWord(){
	//time to next track data word
	return (((m_currentTrackOffset + 2.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2)*8333333.333) - ((m_currentTrackOffset + 0.001)/(m_EncodedBytesPerTrack + 0.001)*(m_Is288RPM ? 0.208333333 : 0.2)*8333333.333));
};


unsigned int WD1772Real::CalculateCRC(unsigned int CRCStart, unsigned char byteToEval){
	unsigned int tmpCRC = CRCStart;
	unsigned char tmpInByte = byteToEval;
	unsigned char tmpByteD1;
	for (int loop = 0; loop < 8; ++loop){
		tmpByteD1 = (tmpByteD1 << 1) + (tmpInByte & 0x01);
		tmpInByte = tmpInByte >> 1;
	};
	tmpCRC ^= tmpByteD1;
	for (int loop = 0; loop < 8; ++loop){
		bool tmpBit = tmpCRC & 0x01;
		tmpCRC = tmpCRC >> 1;
		if(tmpBit){
			tmpCRC ^= 0x8408;
		}
	};
	return tmpCRC;
};

unsigned int WD1772Real::CalculateCRCEnd(unsigned int CRCStart){
	unsigned int tmpCRC = CRCStart;
	unsigned int endCRC = 0;
	for (int loop = 0; loop < 16; ++loop){
		endCRC = (endCRC << 1) + (tmpCRC & 0x01);
		tmpCRC = tmpCRC >> 1;
	};
	return endCRC;
};

void WD1772Real::SetTrackDataWord(unsigned char * trackPointer, unsigned int dataWord){
	trackPointer[0] = (dataWord >> 8) & 0xff;
	trackPointer[1] = dataWord & 0xff;
};

unsigned long int WD1772Real::GetTrackDataLongword(unsigned char * trackStartPointer, unsigned int dataOffset){
	unsigned char * tmpPointer = trackStartPointer + dataOffset;
	unsigned long int tmpData = (tmpPointer[0] << 24) + (tmpPointer[1] << 16) + (tmpPointer[2] << 8) + tmpPointer[3];
	if((m_EncodedBytesPerTrack - dataOffset) <= 2){
		tmpData &= 0xffff0000;
		tmpData += (trackStartPointer[0] << 8) + trackStartPointer[1];
	}

	return tmpData;
};

#ifdef WITH_CAPS
//dataoffset may be up to approx. 2x tracklength
unsigned long int WD1772Real::GetTrackDataLongWordCT(char currentRevolution, unsigned int dataOffset){
	//bigger than second revolution
	if(dataOffset >= (((m_CAPSTrackInfo[currentRevolution].tracklen+m_CAPSTrackInfo[currentRevolution+1].tracklen) >> 3) - 8))
		return -1;
	//in first revolution
	if(dataOffset <= ((m_CAPSTrackInfo[currentRevolution].tracklen >> 3) - 4)){
		unsigned char * tmpPointer = m_CAPSTrackInfo[currentRevolution].trackbuf + dataOffset;
		unsigned long int tmpData =
			(tmpPointer[0] << 24) +
			(tmpPointer[1] << 16) +
			(tmpPointer[2] << 8) +
			tmpPointer[3];
		return tmpData;
	};

	//at boundary to next revolution
	if(dataOffset < (m_CAPSTrackInfo[currentRevolution].tracklen >> 3)){
		unsigned char * tmpPointer = m_CAPSTrackInfo[currentRevolution].trackbuf + dataOffset;
		unsigned char * tmpPointer2 = m_CAPSTrackInfo[currentRevolution+1].trackbuf;
		unsigned long int tmpData = 0;
		switch((m_CAPSTrackInfo[currentRevolution].tracklen >> 3) - dataOffset){
		case 3:
			tmpData = 
				(tmpPointer[0] << 24) +
				(tmpPointer[1] << 16) +
				(tmpPointer[2] << 8) +
				tmpPointer2[0];
			return tmpData;

		case 2:
			tmpData = 
				(tmpPointer[0] << 24) +
				(tmpPointer[1] << 16) +
				(tmpPointer2[0] << 8) +
				tmpPointer2[1];
			return tmpData;

		case 1:
			tmpData = 
				(tmpPointer[0] << 24) +
				(tmpPointer2[0] << 16) +
				(tmpPointer2[1] << 8) +
				tmpPointer2[2];
			return tmpData;

		};
	};
	//in second revolution
	unsigned char * tmpPointer2 = m_CAPSTrackInfo[currentRevolution+1].trackbuf + 
		(dataOffset - (m_CAPSTrackInfo[currentRevolution].tracklen >> 3));
	unsigned long int tmpData = 
		(tmpPointer2[0] << 24) +
		(tmpPointer2[1] << 16) +
		(tmpPointer2[2] << 8) +
		tmpPointer2[3];
	return tmpData;
};

//dataoffset may be up to approx. 2x tracklength
uint64_t WD1772Real::GetTrackDataQuadWordCT(char currentRevolution, unsigned int dataOffset){
	//bigger than second revolution
	if(dataOffset >= (((m_CAPSTrackInfo[currentRevolution].tracklen+m_CAPSTrackInfo[currentRevolution+1].tracklen) >> 3) - 16))
		return -1;
	//in first revolution
	if(dataOffset <= ((m_CAPSTrackInfo[currentRevolution].tracklen >> 3) - 8)){
		unsigned char * tmpPointer = m_CAPSTrackInfo[currentRevolution].trackbuf + dataOffset;
		uint64_t tmpData =
			((uint64_t) tmpPointer[0] << 56) +
			((uint64_t) tmpPointer[1] << 48) +
			((uint64_t) tmpPointer[2] << 40) +
			((uint64_t) tmpPointer[3] << 32) +
			((uint64_t) tmpPointer[4] << 24) +
			((uint64_t) tmpPointer[5] << 16) +
			((uint64_t) tmpPointer[6] << 8) +
			(uint64_t) tmpPointer[7];
		return tmpData;
	};

	//at boundary to next revolution
	if(dataOffset < (m_CAPSTrackInfo[currentRevolution].tracklen >> 3)){
		unsigned char * tmpPointer = m_CAPSTrackInfo[currentRevolution].trackbuf + dataOffset;
		unsigned char * tmpPointer2 = m_CAPSTrackInfo[currentRevolution+1].trackbuf;
		uint64_t tmpData = 0;
		switch((m_CAPSTrackInfo[currentRevolution].tracklen >> 3) - dataOffset){
		case 7:
			tmpData = 
				((uint64_t) tmpPointer[0] << 56) +
				((uint64_t) tmpPointer[1] << 48) +
				((uint64_t) tmpPointer[2] << 40) +
				((uint64_t) tmpPointer[3] << 32) +
				((uint64_t) tmpPointer[4] << 24) +
				((uint64_t) tmpPointer[5] << 16) +
				((uint64_t) tmpPointer[6] << 8) +
				((uint64_t) tmpPointer2[0]);
			return tmpData;

		case 6:
			tmpData = 
				((uint64_t) tmpPointer[0] << 56) +
				((uint64_t) tmpPointer[1] << 48) +
				((uint64_t) tmpPointer[2] << 40) +
				((uint64_t) tmpPointer[3] << 32) +
				((uint64_t) tmpPointer[4] << 24) +
				((uint64_t) tmpPointer[5] << 16) +
				((uint64_t) tmpPointer2[0] << 8) +
				((uint64_t) tmpPointer2[1]);
			return tmpData;

		case 5:
			tmpData = 
				((uint64_t) tmpPointer[0] << 56) +
				((uint64_t) tmpPointer[1] << 48) +
				((uint64_t) tmpPointer[2] << 40) +
				((uint64_t) tmpPointer[3] << 32) +
				((uint64_t) tmpPointer[4] << 24) +
				((uint64_t) tmpPointer2[0] << 16) +
				((uint64_t) tmpPointer2[1] << 8) +
				((uint64_t) tmpPointer2[2]);
			return tmpData;

		case 4:
			tmpData = 
				((uint64_t) tmpPointer[0] << 56) +
				((uint64_t) tmpPointer[1] << 48) +
				((uint64_t) tmpPointer[2] << 40) +
				((uint64_t) tmpPointer[3] << 32) +
				((uint64_t) tmpPointer2[0] << 24) +
				((uint64_t) tmpPointer2[1] << 16) +
				((uint64_t) tmpPointer2[2] << 8) +
				((uint64_t) tmpPointer2[3]);
			return tmpData;

		case 3:
			tmpData = 
				((uint64_t) tmpPointer[0] << 56) +
				((uint64_t) tmpPointer[1] << 48) +
				((uint64_t) tmpPointer[2] << 40) +
				((uint64_t) tmpPointer2[0] << 32) +
				((uint64_t) tmpPointer2[1] << 24) +
				((uint64_t) tmpPointer2[2] << 16) +
				((uint64_t) tmpPointer2[3] << 8) +
				((uint64_t) tmpPointer2[4]);
			return tmpData;

		case 2:
			tmpData = 
				((uint64_t) tmpPointer[0] << 56) +
				((uint64_t) tmpPointer[1] << 48) +
				((uint64_t) tmpPointer2[0] << 40) +
				((uint64_t) tmpPointer2[1] << 32) +
				((uint64_t) tmpPointer2[2] << 24) +
				((uint64_t) tmpPointer2[3] << 16) +
				((uint64_t) tmpPointer2[4] << 8) +
				((uint64_t) tmpPointer2[5]);
			return tmpData;

		case 1:
			tmpData = 
				((uint64_t) tmpPointer[0] << 56) +
				((uint64_t) tmpPointer2[0] << 48) +
				((uint64_t) tmpPointer2[1] << 40) +
				((uint64_t) tmpPointer2[2] << 32) +
				((uint64_t) tmpPointer2[3] << 24) +
				((uint64_t) tmpPointer2[4] << 16) +
				((uint64_t) tmpPointer2[5] << 8) +
				((uint64_t) tmpPointer2[6]);
			return tmpData;

		};
	};
	//in second revolution
	unsigned char * tmpPointer2 = m_CAPSTrackInfo[currentRevolution+1].trackbuf + 
		(dataOffset - (m_CAPSTrackInfo[currentRevolution].tracklen >> 3));
	uint64_t tmpData = 
		((uint64_t) tmpPointer2[0] << 56) +
		((uint64_t) tmpPointer2[1] << 48) +
		((uint64_t) tmpPointer2[2] << 40) +
		((uint64_t) tmpPointer2[3] << 32) +
		((uint64_t) tmpPointer2[4] << 24) +
		((uint64_t) tmpPointer2[5] << 16) +
		((uint64_t) tmpPointer2[6] << 8) +
		((uint64_t) tmpPointer2[7]);
	return tmpData;
};

signed char WD1772Real::SearchSyncsCTFM(uint64_t myDataQuadWord, unsigned long mySync1, unsigned long mySync2, unsigned long mySync3){
	uint64_t tmpData = myDataQuadWord;
	signed char bitshift = 31;
	while(bitshift > -1){
		tmpData = (myDataQuadWord >> (32 - bitshift)) & 0xffffffff;
		if((tmpData == mySync1) || (tmpData == mySync2) || (tmpData == mySync3)){
#ifdef TESTSTATE
			printf ("Sync CTFM detected\n");
#endif
			break;
		};
		--bitshift;
	};
	return bitshift;

};
#endif

signed char WD1772Real::SearchSync(unsigned long int myDataLongWord, unsigned int mySync1, unsigned int mySync2){
	unsigned long int tmpData = myDataLongWord;
	signed char bitshift = 15;
	while(bitshift > -1){
		tmpData = (myDataLongWord >> (16 - bitshift)) & 0xffff;
		if((tmpData  == mySync1) || (tmpData == mySync2)){
#ifdef TESTSTATE
			printf ("Sync detected\n");
#endif
			break;
		};
		--bitshift;
	};
	return bitshift;
};

void WD1772Real::ResetTrackWriteProperties(bool noData){
	for(int iloop = 0; iloop < 26; ++iloop){
		for(int jloop = 0; jloop < 30; ++jloop){
			m_SectorLayout[iloop][jloop].OffsetInProcessed = -1;
			m_SectorLayout[iloop][jloop].SectorSequenceNumber = 0xff;
			if(!noData && (m_SectorLayout[iloop][jloop].SectorData != NULL)){
				delete m_SectorLayout[iloop][jloop].SectorData;
			};
			m_SectorLayout[iloop][jloop].SectorData = NULL;
		};
	};
	m_SectorLengthInTrack = 0x100;
};

void WD1772Real::TransferTrackWriteProperties(){
	char side = m_IsDiskSide1 ? 1 : 0;
	if(m_WATRHeader.SidePtr[side] != NULL){
		if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack] != NULL){
			for(int iloop = 0; iloop < 30; ++iloop){
				if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop] != NULL){
					m_SectorLayout[m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorNumber][m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorSequence].SectorSequenceNumber = m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorSequence;
					m_SectorLayout[m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorNumber][m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorSequence].OffsetInProcessed = m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorOffset;
					if(m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorData != NULL){
						m_SectorLayout[m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorNumber][m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorSequence].SectorData = new unsigned char[m_WATRHeader.DiskSectorLen];
						memcpy(m_SectorLayout[m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorNumber][m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorSequence].SectorData, m_WATRHeader.SidePtr[side]->TrackPtr[m_CurrentTrack]->SectorPtr[iloop]->SectorData, m_WATRHeader.DiskSectorLen);
					};
				};
			};
		};
	};
};

void WD1772Real::FreeWATRData(){
	for(int iloop = 0; iloop < 2; ++iloop){
		if(m_WATRHeader.SidePtr[iloop] != NULL){
			for(int jloop = 0; jloop < 40; ++jloop){
				if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop] != NULL){
					for(int kloop = 0; kloop < 30; ++kloop){
						if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop] != NULL){
							if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorData != NULL)
								delete m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorData;
							delete m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop];
						};
						m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop] = NULL;
					};
					delete m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop];
				};
				m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop] = NULL;
			};
			delete m_WATRHeader.SidePtr[iloop];
		};
		m_WATRHeader.SidePtr[iloop] = NULL;
	};
};

void WD1772Real::LoadWATRData(){
	m_WATRHeader.SidePtr[0] = NULL;
	m_WATRHeader.SidePtr[1] = NULL;
	m_WATRHeader.DiskSectorLen = 0x100;

	//WATR file structure:
	//to reflect Header structure
	//4 bytes "WATR" at offset 0
	//2 bytes version hi, lo
	//2 bytes disk sector len hi, lo
	//2x4 bytes signed long int offset of WATRSide in file hi, hm, lm, lo

	//to reflect Side Structure
	//40x4 bytes signed long int offset of WATRTrack in file hi, hm, lm, lo

	//to reflect Track Structure
	//30x4 bytes signed long int offset of WATRSectors in file hi, hm, lm, lo

	//to reflect Sector Structure
	//1 byte Sectornumber
	//1 byte sector sequence number
	//2 bytes unsigned short int offset of sector in encoded data hi, lo
	//4 bytes signed long int offset of sector data in file hi, hm, lm, lo

	FILE * f;
	f = fopen(&m_WATRFileName[0], "rb");
	if (f == NULL)
		return;

	//compare Identifier
	if(fread(&m_WATRHeader.W_ID[0], 1, 4, f) != 4){
		fclose(f);
		return;
	}
	if(strncmp(&m_WATRHeader.W_ID[0], "WATR", 4) != 0){
		fclose(f);
		return;
	};

	//compare WATR-File-Version
	unsigned char tmpVersion[2];
	if(fread(&tmpVersion[0], 1, 2, f)!= 2){
		fclose(f);
		return;
	};
	unsigned short int tmpVer = (tmpVersion[0] << 8) + (tmpVersion[1]);
	if(tmpVer != 1){
		fclose(f);
		return;
	};

	//get disk sector length
	unsigned char tmpDSLen[2];
	if(fread(&tmpDSLen[0], 1, 2, f)!= 2){
		fclose(f);
		return;
	};
	unsigned short int tmpDSLeng = (tmpDSLen[0] << 8) + tmpDSLen[1];
	m_WATRHeader.DiskSectorLen = tmpDSLeng;

#ifdef TESTSTATE
	printf("WATR headers loaded.\n");
#endif
	//now load structures
	unsigned char tmpBuf[4];
	for(int iloop = 0; iloop < 2; ++iloop){
		if(fread(&tmpBuf[0], 1, 4, f) != 4){
			fclose(f);
			FreeWATRData();
			return;
		};
		unsigned long int tmpSide = (tmpBuf[0] << 24) + (tmpBuf[1] << 16) + (tmpBuf[2] << 8) + tmpBuf[3];
#ifdef TESTSTATE
	printf("Side Offset: %8x\n",tmpSide);
#endif
		if(tmpSide != -1){
			m_WATRHeader.SidePtr[iloop] = new WATRSide;
			for(int zloop = 0; zloop < 40; ++zloop){
				m_WATRHeader.SidePtr[iloop]->TrackPtr[zloop] = NULL;
			};
			long int fposHeader = ftell(f);
			fseek(f, tmpSide, SEEK_SET);
			for(int jloop = 0; jloop < 40; ++jloop){
				if(fread(&tmpBuf[0], 1, 4, f) != 4){
					fclose(f);
					FreeWATRData();
					return;
				};
				unsigned long int tmpTrack = (tmpBuf[0] << 24) + (tmpBuf[1] << 16) + (tmpBuf[2] << 8) + tmpBuf[3];
#ifdef TESTSTATE
	printf("Track Offset: %8x\n",tmpTrack);
#endif
				if(tmpTrack != -1){
					m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop] = new WATRTrack;
					for(int zloop = 0; zloop < 30; ++zloop){
						m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[zloop] = NULL;
					};
					long int fposSide = ftell(f);
					fseek(f, tmpTrack, SEEK_SET);
					for(int kloop = 0; kloop < 30; ++kloop){					
						if(fread(&tmpBuf[0], 1, 4, f) != 4){
							fclose(f);
							FreeWATRData();
							return;
						};
						unsigned long int tmpSector = (tmpBuf[0] << 24) + (tmpBuf[1] << 16) + (tmpBuf[2] << 8) + tmpBuf[3];
#ifdef TESTSTATE
	printf("Sector Offset: %8x\n",tmpSector);
#endif
						if(tmpSector != -1){
							m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop] = new WATRSectors;
							m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorData = NULL;
							long int fposTrack = ftell(f);
							fseek(f, tmpSector, SEEK_SET);
							//read sectornumber
							if(fread(&tmpBuf[0], 1, 1, f) != 1){
								fclose(f);
								FreeWATRData();
								return;
							};
							m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorNumber = tmpBuf[0];
							//read sequence number
							if(fread(&tmpBuf[0], 1, 1, f) != 1){
								fclose(f);
								FreeWATRData();
								return;
							};
							m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorSequence = tmpBuf[0];
							//read offset
							if(fread(&tmpBuf[0], 1, 2, f) != 2){
								fclose(f);
								FreeWATRData();
								return;
							};
							m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorOffset = (tmpBuf[0] << 8) + tmpBuf[1];
							//read data file offset
							if(fread(&tmpBuf[0], 1, 4, f) != 4){
								fclose(f);
								FreeWATRData();
								return;
							};
							unsigned long int tmpSectorData = (tmpBuf[0] << 24) + (tmpBuf[1] << 16) + (tmpBuf[2] << 8) + tmpBuf[3];
#ifdef TESTSTATE
	printf("Sector Data Offset: %8x\n",tmpSectorData);
#endif
							if(tmpSectorData != -1){
								m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorData = new unsigned char[m_WATRHeader.DiskSectorLen];
								fseek(f, tmpSectorData, SEEK_SET);
								if(fread(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]
									->SectorData, 1, m_WATRHeader.DiskSectorLen, f) 
									!= m_WATRHeader.DiskSectorLen){
									fclose(f);
									FreeWATRData();
									return;
								};
							};
							fseek(f, fposTrack, SEEK_SET);
						};
					};
					fseek(f, fposSide, SEEK_SET);
				};
			};
			fseek(f, fposHeader, SEEK_SET);
		};
	};
	fclose(f);
#ifdef TESTSTATE
	printf("WATR file loaded.\n");
#endif
};

void WD1772Real::SaveWATRData(){

	if((m_WATRHeader.SidePtr[0] == NULL) && (m_WATRHeader.SidePtr[1] == NULL))
		return;

	//calculate data offsets
	//sizeof WATRHeader in file is 16 bytes
	//sizeof WATRSide in file is 160 bytes
	//sizeof WATRTrack in file is 120 bytes
	//sizeof WATRSector in file is 8 bytes
	//sizeof SectorData in file is m_WATRHeader.DiskSectorLen
	unsigned long int tmpWSideSize = 0;
	unsigned long int tmpWTrackSize = 0;
	unsigned long int tmpWSectorSize = 0;
	unsigned long int tmpWDataSize = 0;
	for(int iloop = 0; iloop < 2; ++iloop){
		if(m_WATRHeader.SidePtr[iloop] != NULL){
			m_WATRHeader.SidePtr[iloop]->FileOffset = tmpWSideSize;
			tmpWSideSize += 160;
			for(int jloop = 0; jloop < 40; ++jloop){
				if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop] != NULL){
					m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->FileOffset = tmpWTrackSize;
					tmpWTrackSize +=120;
					for(int kloop = 0; kloop < 30; ++kloop){
						if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop] != NULL){
							m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->FileOffsetSectorDescr = tmpWSectorSize;
							tmpWSectorSize += 8;
							if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorData != NULL){
								m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->FileOffsetSectorData = tmpWDataSize;
								tmpWDataSize += m_WATRHeader.DiskSectorLen;
							};
						};
					};
				};
			};
		};
	};

	//now save:
	FILE * f;
	f = fopen(&m_WATRFileName[0], "wb+");
#ifdef TESTSTATE
	if(f != NULL){
		printf("Open for Writing succeeded.\n");
	}
	else{
		printf("Open for Writing failed.\n");
	};	
#endif
	if(f == NULL)
		return;	

	//the WATR header
	char tmpWATRstr[5] = "WATR";
	fwrite(&tmpWATRstr[0], 1, 4, f);

	unsigned char tmpBuf[4];
	tmpBuf[0] = 0;
	tmpBuf[1] = 1;
	fwrite(&tmpBuf[0], 1, 2, f);

	tmpBuf[0] = (m_WATRHeader.DiskSectorLen >> 8) & 0xff;
	tmpBuf[1] = m_WATRHeader.DiskSectorLen & 0xff;
	fwrite(&tmpBuf[0], 1, 2, f);

	for(int iloop = 0; iloop < 2; ++iloop){
		if(m_WATRHeader.SidePtr[iloop] != NULL){
			unsigned long int tmpSideOffset = m_WATRHeader.SidePtr[iloop]->FileOffset + 16;
			tmpBuf[0] = (tmpSideOffset >> 24) & 0xff;
			tmpBuf[1] = (tmpSideOffset >> 16) & 0xff;
			tmpBuf[2] = (tmpSideOffset >> 8) & 0xff;
			tmpBuf[3] = tmpSideOffset & 0xff;
			fwrite(&tmpBuf[0], 1, 4, f);
		}
		else{
			tmpBuf[0] = 0xff;
			tmpBuf[1] = 0xff;
			tmpBuf[2] = 0xff;
			tmpBuf[3] = 0xff;
			fwrite(&tmpBuf[0], 1, 4, f);
		};
	};

	//save the sides
	for(int iloop = 0; iloop < 2; ++iloop){
		if(m_WATRHeader.SidePtr[iloop] != NULL){
			for(int jloop = 0; jloop < 40; ++jloop){
				if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop] != NULL){
					unsigned long int tmpTrackOffset = m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->FileOffset + 16 + tmpWSideSize;
					tmpBuf[0] = (tmpTrackOffset >> 24) & 0xff;
					tmpBuf[1] = (tmpTrackOffset >> 16) & 0xff;
					tmpBuf[2] = (tmpTrackOffset >> 8) & 0xff;
					tmpBuf[3] = tmpTrackOffset & 0xff;
					fwrite(&tmpBuf[0], 1, 4, f);
				}
				else{
					tmpBuf[0] = 0xff;
					tmpBuf[1] = 0xff;
					tmpBuf[2] = 0xff;
					tmpBuf[3] = 0xff;
					fwrite(&tmpBuf[0], 1, 4, f);
				};
			};
		};
	};

	//save the tracks
	for(int iloop = 0; iloop < 2; ++iloop){
		if(m_WATRHeader.SidePtr[iloop] != NULL){
			for(int jloop = 0; jloop < 40; ++jloop){
				if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop] != NULL){
					for(int kloop = 0; kloop < 30; ++kloop){
						if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop] != NULL){
							unsigned long int tmpSectorOffset = m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->FileOffsetSectorDescr + 16 + tmpWSideSize + tmpWTrackSize;
							tmpBuf[0] = (tmpSectorOffset >> 24) & 0xff;
							tmpBuf[1] = (tmpSectorOffset >> 16) & 0xff;
							tmpBuf[2] = (tmpSectorOffset >> 8) & 0xff;
							tmpBuf[3] = tmpSectorOffset & 0xff;
							fwrite(&tmpBuf[0], 1, 4, f);
						}
						else{
							tmpBuf[0] = 0xff;
							tmpBuf[1] = 0xff;
							tmpBuf[2] = 0xff;
							tmpBuf[3] = 0xff;
							fwrite(&tmpBuf[0], 1, 4, f);
						};
					};
				};
			};
		};
	};
	//save the sector descriptions
	for(int iloop = 0; iloop < 2; ++iloop){
		if(m_WATRHeader.SidePtr[iloop] != NULL){
			for(int jloop = 0; jloop < 40; ++jloop){
				if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop] != NULL){
					for(int kloop = 0; kloop < 30; ++kloop){
						if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop] != NULL){
							//save sector number
							fwrite(&m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorNumber, 1, 1, f);
							//save sector sequence number
							fwrite(&m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorSequence, 1, 1, f);
							//save sector offset in raw data
							unsigned short int tmpOffs = m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorOffset;
							tmpBuf[0] = (tmpOffs >> 8) & 0xff;
							tmpBuf[1] = tmpOffs & 0xff;
							fwrite(&tmpBuf[0], 1, 2, f);

							//save sector data ptr
							if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorData != NULL){
								unsigned long int tmpSectorDataOffset = m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->FileOffsetSectorData + 16 + tmpWSideSize + tmpWTrackSize + tmpWSectorSize;
								tmpBuf[0] = (tmpSectorDataOffset >> 24) & 0xff;
								tmpBuf[1] = (tmpSectorDataOffset >> 16) & 0xff;
								tmpBuf[2] = (tmpSectorDataOffset >> 8) & 0xff;
								tmpBuf[3] = tmpSectorDataOffset & 0xff;
								fwrite(&tmpBuf[0], 1, 4, f);
							}
							else{
								tmpBuf[0] = 0xff;
								tmpBuf[1] = 0xff;
								tmpBuf[2] = 0xff;
								tmpBuf[3] = 0xff;
								fwrite(&tmpBuf[0], 1, 4, f);
							};
						};
					};
				};
			};
		};
	};

	//save the sector data
	for(int iloop = 0; iloop < 2; ++iloop){
		if(m_WATRHeader.SidePtr[iloop] != NULL){
			for(int jloop = 0; jloop < 40; ++jloop){
				if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop] != NULL){
					for(int kloop = 0; kloop < 30; ++kloop){
						if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop] != NULL){
							if(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorData != NULL){
								fwrite(m_WATRHeader.SidePtr[iloop]->TrackPtr[jloop]->SectorPtr[kloop]->SectorData, 1, m_WATRHeader.DiskSectorLen, f);
							};
						};
					};
				};
			};
		};
	};

	fclose(f);
};

}

/*
vim:ts=4:sw=4:
*/
